from pyRDDLGym.Core.Parser import parser as parser
from pyRDDLGym.Core.Parser import RDDLReader as RDDLReader
from pyRDDLGym.Core import Grounder as RDDLGrounder
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulator

DOMAIN = 'rddl_termination_test.rddl'

MyReader = RDDLReader.RDDLReader('RDDL/' + DOMAIN)
domain = MyReader.rddltxt

MyLexer = parser.RDDLlex()
MyLexer.build()
MyLexer.input(domain)
token_list = [token for token in MyLexer._lexer]
# print(token_list)

# build parser - built in lexer, non verbose
MyRDDLParser = parser.RDDLParser(None, False)
MyRDDLParser.build()

# parse RDDL file
rddl_ast = MyRDDLParser.parse(domain)
grounder = RDDLGrounder.RDDLGrounder(rddl_ast)
model = grounder.Ground()
sampler = RDDLSimulator(model)

loops = 2
for h in range(loops):
    state = sampler.reset_state()
    print(state)
    action = {
        'curProd_p1': 1,
        'curProd_p2': 2,
        'curProd_p3': 3,
    }
    state = sampler.sample_next_state(action)
    sampler.check_terminal_states()
    print(state)

print ('hello world')


'''In this example, the user has the choice to run the Jax planner with three
different options:
    
    1. slp runs the straight-line planner offline, which trains an open-loop plan
    2. drp runs the deep reactive policy, which trains a policy network
    3. replan runs the straight-line planner online, at every decision epoch
    
The syntax for running this example is:

    python JaxExample.py <domain> <instance> <method>
    
where:
    <domain> is the name of a domain located in the /Examples directory
    <instance> is the instance number
    <method> is either slp, drp, or replan
'''
import os
import sys

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import load_config
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxRDDLBackpropPlanner
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxOfflineController
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxOnlineController
from pyRDDLGym.Examples.ExampleManager import ExampleManager

    
def main(domain, instance, method):
    
    # set up the environment
    info = ExampleManager.GetEnvInfo(domain)    
    env = RDDLEnv.build(info, instance, enforce_action_constraints=True)
    
    # load the config file with planner settings
    abs_path = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(abs_path, 'JaxPlanConfigs', f'{domain}_{method}.cfg') 
    planner_args, _, train_args = load_config(config_path)
    
    # create the planning algorithm
    planner = JaxRDDLBackpropPlanner(rddl=env.model, **planner_args)
    
    # create the controller   
    if method == 'replan':
        controller = JaxOnlineController(planner, **train_args)
    else:
        controller = JaxOfflineController(planner, **train_args)
    
    controller.evaluate(env, verbose=True, render=True)
    
    env.close()
        
        
if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 3:
        print('python JaxExample.py <domain> <instance> <method>')
        exit(1)
    if args[2] not in ['drp', 'slp', 'replan']:
        print('<method> in [drp, slp, replan]')
        exit(1)
    domain, instance, method = args[:3]
    main(domain, instance, method)
    



import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class HVACInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return 'HVAC'
    
    def get_domain_name(self) -> str:
        return 'hvac'
    
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        nh = params['num_heaters']
        nz = params['num_zones']
        heaters, zones = self._generate_layout(nh, nz, params['density'])
        
        obj_heaters = [f'h{i + 1}' for i in range(nh)]
        obj_zones = [f'z{i + 1}' for i in range(nz)]
        
        nonfluents = {}
        nonfluents['TEMP-ZONE-MIN'] = params['TEMP-ZONE-MIN']
        nonfluents['TEMP-ZONE-MAX'] = params['TEMP-ZONE-MAX']
        zones_to_switch = list(range(nz))
        random.shuffle(zones_to_switch)
        for i in zones_to_switch[:params['p-switch-number']]:
            z = obj_zones[i]
            nonfluents[f'P-SWITCH({z})'] = params['p-switch-prob']
        for h, h2z in zip(obj_heaters, heaters):
            for z in h2z:
                nonfluents[f'ADJ-HEATER({h}, {obj_zones[z]})'] = True
        for z, z2z in zip(obj_zones, zones):
            for z2 in z2z:
                nonfluents[f'ADJ-ZONES({z}, {obj_zones[z2]})'] = True
        
        states = {}
        for i, z in enumerate(obj_zones):
            states[f'temp-zone({z})'] = random.uniform(*params['temp-zone-range-init'])
        for i, h in enumerate(obj_heaters):
            states[f'temp-heater({h})'] = random.uniform(*params['temp-heater-range-init'])
                
        return {
            'objects': {'zone': obj_zones, 'heater': obj_heaters},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }
            
    def _generate_layout(self, nh, nz, density):
        heaters = [set() for _ in range(nh)]
        zones = [set() for _ in range(nz)]
    
        # Each heater must be connected to at least one zone
        for h in range(nh):
            z = random.randrange(nz)
            heaters[h].add(z)
    
        # Each zone must be connected to at least one heater
        for z in range(nz):
            if not any((z in htz) for htz in heaters):
                h = random.randrange(nh)
                heaters[h].add(z)
    
        # Zones can be interconnected
        for z1 in range(nz):
            for z2 in range(z1 + 1, nz):
                if random.random() < density:
                    zones[z1].add(z2)
        return heaters, zones
    



from functools import partial
import jax
import jax.numpy as jnp
import jax.random as random
import jax.scipy as scipy 
import traceback
from typing import Callable, Dict, List
import warnings

# more robust approach - if user does not have this or broken try to continue
try:
    from tensorflow_probability.substrates import jax as tfp
except Exception:
    warnings.warn('Failed to import tensorflow-probability: '
                  'compilation of some complex distributions will not work.',
                  stacklevel=2)
    traceback.print_exc()
    tfp = None
    
from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError

from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Compiler.RDDLObjectsTracer import RDDLObjectsTracer
from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
from pyRDDLGym.Core.Debug.Logger import Logger
from pyRDDLGym.Core.Env.RDDLConstraints import RDDLConstraints
from pyRDDLGym.Core.Simulator.RDDLSimulator import RDDLSimulatorPrecompiled


class JaxRDDLCompiler:
    '''Compiles a RDDL AST representation into an equivalent JAX representation.
    All operations are identical to their numpy equivalents.
    '''
    
    MODEL_PARAM_TAG_SEPARATOR = '___'
    
    def __init__(self, rddl: RDDLLiftedModel,
                 allow_synchronous_state: bool=True,
                 logger: Logger=None,
                 use64bit: bool=False) -> None:
        '''Creates a new RDDL to Jax compiler.
        
        :param rddl: the RDDL model to compile into Jax
        :param allow_synchronous_state: whether next-state components can depend
        on each other
        :param logger: to log information about compilation to file
        :param use64bit: whether to use 64 bit arithmetic
        '''
        self.rddl = rddl
        self.logger = logger
        # jax.config.update('jax_log_compiles', True) # for testing ONLY
        
        if use64bit:
            self.INT = jnp.int64
            self.REAL = jnp.float64
            jax.config.update('jax_enable_x64', True)
        else:
            self.INT = jnp.int32
            self.REAL = jnp.float32
        self.ONE = jnp.asarray(1, dtype=self.INT)
        self.JAX_TYPES = {
            'int': self.INT,
            'real': self.REAL,
            'bool': bool
        }
        
        # compile initial values
        if self.logger is not None:
            self.logger.clear()
        initializer = RDDLValueInitializer(rddl, logger=self.logger)
        self.init_values = initializer.initialize()
        
        # compute dependency graph for CPFs and sort them by evaluation order
        sorter = RDDLLevelAnalysis(rddl, allow_synchronous_state, logger=self.logger)
        self.levels = sorter.compute_levels()        
        
        # trace expressions to cache information to be used later
        tracer = RDDLObjectsTracer(rddl, logger=self.logger, cpf_levels=self.levels)
        self.traced = tracer.trace()
        
        # extract the box constraints on actions
        simulator = RDDLSimulatorPrecompiled(
            rddl=self.rddl,
            init_values=self.init_values,
            levels=self.levels,
            trace_info=self.traced)  
        constraints = RDDLConstraints(simulator, vectorized=True)
        self.constraints = constraints
        
        # basic operations
        self.NEGATIVE = lambda x, param: jnp.negative(x)  
        self.ARITHMETIC_OPS = {
            '+': lambda x, y, param: jnp.add(x, y),
            '-': lambda x, y, param: jnp.subtract(x, y),
            '*': lambda x, y, param: jnp.multiply(x, y),
            '/': lambda x, y, param: jnp.divide(x, y)
        }    
        self.RELATIONAL_OPS = {
            '>=': lambda x, y, param: jnp.greater_equal(x, y),
            '<=': lambda x, y, param: jnp.less_equal(x, y),
            '<': lambda x, y, param: jnp.less(x, y),
            '>': lambda x, y, param: jnp.greater(x, y),
            '==': lambda x, y, param: jnp.equal(x, y),
            '~=': lambda x, y, param: jnp.not_equal(x, y)
        }
        self.LOGICAL_NOT = lambda x, param: jnp.logical_not(x)
        self.LOGICAL_OPS = {
            '^': lambda x, y, param: jnp.logical_and(x, y),
            '&': lambda x, y, param: jnp.logical_and(x, y),
            '|': lambda x, y, param: jnp.logical_or(x, y),
            '~': lambda x, y, param: jnp.logical_xor(x, y),
            '=>': lambda x, y, param: jnp.logical_or(jnp.logical_not(x), y),
            '<=>': lambda x, y, param: jnp.equal(x, y)
        }
        self.AGGREGATION_OPS = {
            'sum': lambda x, axis, param: jnp.sum(x, axis=axis),
            'avg': lambda x, axis, param: jnp.mean(x, axis=axis),
            'prod': lambda x, axis, param: jnp.prod(x, axis=axis),
            'minimum': lambda x, axis, param: jnp.min(x, axis=axis),
            'maximum': lambda x, axis, param: jnp.max(x, axis=axis),
            'forall': lambda x, axis, param: jnp.all(x, axis=axis),
            'exists': lambda x, axis, param: jnp.any(x, axis=axis),
            'argmin': lambda x, axis, param: jnp.argmin(x, axis=axis),
            'argmax': lambda x, axis, param: jnp.argmax(x, axis=axis)
        }
        self.AGGREGATION_BOOL = {'forall', 'exists'}
        self.KNOWN_UNARY = {        
            'abs': lambda x, param: jnp.abs(x),
            'sgn': lambda x, param: jnp.sign(x),
            'round': lambda x, param: jnp.round(x),
            'floor': lambda x, param: jnp.floor(x),
            'ceil': lambda x, param: jnp.ceil(x),
            'cos': lambda x, param: jnp.cos(x),
            'sin': lambda x, param: jnp.sin(x),
            'tan': lambda x, param: jnp.tan(x),
            'acos': lambda x, param: jnp.arccos(x),
            'asin': lambda x, param: jnp.arcsin(x),
            'atan': lambda x, param: jnp.arctan(x),
            'cosh': lambda x, param: jnp.cosh(x),
            'sinh': lambda x, param: jnp.sinh(x),
            'tanh': lambda x, param: jnp.tanh(x),
            'exp': lambda x, param: jnp.exp(x),
            'ln': lambda x, param: jnp.log(x),
            'sqrt': lambda x, param: jnp.sqrt(x),
            'lngamma': lambda x, param: scipy.special.gammaln(x),
            'gamma': lambda x, param: jnp.exp(scipy.special.gammaln(x))
        }        
        self.KNOWN_BINARY = {
            'div': lambda x, y, param: jnp.floor_divide(x, y),
            'mod': lambda x, y, param: jnp.mod(x, y),
            'fmod': lambda x, y, param: jnp.mod(x, y),
            'min': lambda x, y, param: jnp.minimum(x, y),
            'max': lambda x, y, param: jnp.maximum(x, y),
            'pow': lambda x, y, param: jnp.power(x, y),
            'log': lambda x, y, param: jnp.log(x) / jnp.log(y),
            'hypot': lambda x, y, param: jnp.hypot(x, y)
        }
        
    # ===========================================================================
    # main compilation subroutines
    # ===========================================================================
     
    def compile(self, log_jax_expr: bool=False) -> None: 
        '''Compiles the current RDDL into Jax expressions.
        
        :param log_jax_expr: whether to pretty-print the compiled Jax functions
        to the log file
        '''
        info = {}
        self.invariants = self._compile_constraints(self.rddl.invariants, info)
        self.preconditions = self._compile_constraints(self.rddl.preconditions, info)
        self.termination = self._compile_constraints(self.rddl.terminals, info)
        self.cpfs = self._compile_cpfs(info)
        self.reward = self._compile_reward(info)
        self.model_params = info
        
        if log_jax_expr and self.logger is not None:
            printed = self.print_jax()
            printed_cpfs = '\n\n'.join(f'{k}: {v}' 
                                       for (k, v) in printed['cpfs'].items())
            printed_reward = printed['reward']
            printed_invariants = '\n\n'.join(v for v in printed['invariants'])
            printed_preconds = '\n\n'.join(v for v in printed['preconditions'])
            printed_terminals = '\n\n'.join(v for v in printed['terminations'])
            printed_params = '\n'.join(f'{k}: {v}' for (k, v) in info.items())
            message = (
                f'[info] compiled JAX CPFs:\n\n'
                f'{printed_cpfs}\n\n'
                f'[info] compiled JAX reward:\n\n'
                f'{printed_reward}\n\n'
                f'[info] compiled JAX invariants:\n\n'
                f'{printed_invariants}\n\n'
                f'[info] compiled JAX preconditions:\n\n'
                f'{printed_preconds}\n\n'
                f'[info] compiled JAX terminations:\n\n'
                f'{printed_terminals}\n'
                f'[info] model parameters:\n'
                f'{printed_params}\n'
            )
            self.logger.log(message)
    
    def _compile_constraints(self, constraints, info):
        return [self._jax(expr, info, dtype=bool) for expr in constraints]
        
    def _compile_cpfs(self, info):
        jax_cpfs = {}
        for cpfs in self.levels.values():
            for cpf in cpfs:
                _, expr = self.rddl.cpfs[cpf]
                prange = self.rddl.variable_ranges[cpf]
                dtype = self.JAX_TYPES.get(prange, self.INT)
                jax_cpfs[cpf] = self._jax(expr, info, dtype=dtype)
        return jax_cpfs
    
    def _compile_reward(self, info):
        return self._jax(self.rddl.reward, info, dtype=self.REAL)
    
    def _extract_inequality_constraint(self, expr):
        result = []
        etype, op = expr.etype
        if etype == 'relational':
            left, right = expr.args
            if op == '<' or op == '<=':
                result.append((left, right))
            elif op == '>' or op == '>=':
                result.append((right, left))
        elif etype == 'boolean' and op == '^':
            for arg in expr.args:
                result.extend(self._extract_inequality_constraint(arg))
        return result
    
    def _extract_equality_constraint(self, expr):
        result = []
        etype, op = expr.etype
        if etype == 'relational':
            left, right = expr.args
            if op == '==':
                result.append((left, right))
        elif etype == 'boolean' and op == '^':
            for arg in expr.args:
                result.extend(self._extract_equality_constraint(arg))
        return result
            
    def _jax_nonlinear_constraints(self): 
        rddl = self.rddl
        
        # extract the non-box inequality constraints on actions
        inequalities = [constr 
                        for (i, expr) in enumerate(rddl.preconditions)
                        for constr in self._extract_inequality_constraint(expr)
                        if not self.constraints.is_box_preconditions[i]]
        
        # compile them to JAX and write as h(s, a) <= 0
        op = self.ARITHMETIC_OPS['-']        
        jax_inequalities = []
        for (left, right) in inequalities:
            jax_lhs = self._jax(left, {})
            jax_rhs = self._jax(right, {})
            jax_constr = self._jax_binary(jax_lhs, jax_rhs, op, '', at_least_int=True)
            jax_inequalities.append(jax_constr)
        
        # extract the non-box equality constraints on actions
        equalities = [constr 
                      for (i, expr) in enumerate(rddl.preconditions)
                      for constr in self._extract_equality_constraint(expr)
                      if not self.constraints.is_box_preconditions[i]]
        
        # compile them to JAX and write as g(s, a) == 0
        jax_equalities = []
        for (left, right) in equalities:
            jax_lhs = self._jax(left, {})
            jax_rhs = self._jax(right, {})
            jax_constr = self._jax_binary(jax_lhs, jax_rhs, op, '', at_least_int=True)
            jax_equalities.append(jax_constr)
            
        return jax_inequalities, jax_equalities
    
    def compile_rollouts(self, policy: Callable,
                         n_steps: int, 
                         n_batch: int,
                         check_constraints: bool=False,
                         constraint_func: bool=False):
        '''Compiles the current RDDL into a wrapped function that samples multiple
        rollouts (state trajectories) in batched form for the given policy. The
        wrapped function takes the policy parameters and RNG key as input, and
        returns a dictionary of all logged information from the rollouts.
        
        constraint_func provides the option to compile nonlinear constraints:
        
            1. f(s, a) ?? g(s, a)
            2. f1(s, a) ^ f2(s, a) ^ ... ?? g(s, a)
            3. forall_{?p1, ...} f(s, a, ?p1, ...) ?? g(s, a) where f is of the
               form 1 or 2 above.
        
        and where ?? is <, <=, > or >= into JAX expressions h(s, a) representing 
        the constraints of the form: 
            
            h(s, a) <= 0
            g(s, a) == 0
                
        for which a penalty or barrier-type method can be used to enforce 
        constraint satisfaction. A list is returned containing values for all
        non-box inequality constraints.
        
        :param policy: a Jax compiled function that takes the policy parameters, 
        decision epoch, state dict, and an RNG key and returns an action dict
        :param n_steps: the length of each rollout
        :param n_batch: how many rollouts each batch performs
        :param check_constraints: whether state, action and termination 
        conditions should be checked on each time step: this info is stored in the
        returned log and does not raise an exception
        :param constraint_func: produces the h(s, a) function described above
        in addition to the usual outputs
        '''
        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
        
        rddl = self.rddl
        reward_fn, cpfs = self.reward, self.cpfs
        preconds, invariants, terminals = \
            self.preconditions, self.invariants, self.termination
            
        if constraint_func:
            inequality_fns, equality_fns = self._jax_nonlinear_constraints()
        else:
            inequality_fns, equality_fns = None, None
        
        # do a single step update from the RDDL model
        def _jax_wrapped_single_step(key, policy_params, hyperparams, 
                                     step, subs, model_params):
            errors = NORMAL
            
            # compute action
            key, subkey = random.split(key)
            states = {var: values 
                      for (var, values) in subs.items()
                      if rddl.variable_types[var] == 'state-fluent'}
            actions = policy(subkey, policy_params, hyperparams, step, states)
            subs.update(actions)
            
            # check action preconditions
            precond_check = True
            if check_constraints:
                for precond in preconds:
                    sample, key, err = precond(subs, model_params, key)
                    precond_check = jnp.logical_and(precond_check, sample)
                    errors |= err
            
            # compute h(s, a) <= 0 and g(s, a) == 0 constraint functions
            inequalities, equalities = [], []
            if constraint_func:
                for constraint in inequality_fns:
                    sample, key, err = constraint(subs, model_params, key)
                    inequalities.append(sample)
                    errors |= err
                for constraint in equality_fns:
                    sample, key, err = constraint(subs, model_params, key)
                    equalities.append(sample)
                    errors |= err
                
            # calculate CPFs in topological order
            for (name, cpf) in cpfs.items():
                subs[name], key, err = cpf(subs, model_params, key)
                errors |= err                
                
            # calculate the immediate reward
            reward, key, err = reward_fn(subs, model_params, key)
            errors |= err
            
            # set the next state to the current state
            for (state, next_state) in rddl.next_state.items():
                subs[state] = subs[next_state]
            
            # check the state invariants
            invariant_check = True
            if check_constraints:
                for invariant in invariants:
                    sample, key, err = invariant(subs, model_params, key)
                    invariant_check = jnp.logical_and(invariant_check, sample)
                    errors |= err
            
            # check the termination (TODO: zero out reward in s if terminated)
            terminated_check = False
            if check_constraints:
                for terminal in terminals:
                    sample, key, err = terminal(subs, model_params, key)
                    terminated_check = jnp.logical_or(terminated_check, sample)
                    errors |= err
            
            # prepare the return value
            log = {
                'pvar': subs,
                'action': actions,
                'reward': reward,
                'error': errors,
                'precondition': precond_check,
                'invariant': invariant_check,
                'termination': terminated_check
            }            
            if constraint_func:
                log['inequalities'] = inequalities
                log['equalities'] = equalities
                
            return log, subs
        
        # do a batched step update from the RDDL model
        def _jax_wrapped_batched_step(carry, step):
            key, policy_params, hyperparams, subs, model_params = carry  
            key, *subkeys = random.split(key, num=1 + n_batch)
            keys = jnp.asarray(subkeys)
            batched_step = jax.vmap(
                _jax_wrapped_single_step,
                in_axes=(0, None, None, None, 0, None)
            ) 
            log, subs = batched_step(
                keys, policy_params, hyperparams, step, subs,  model_params)            
            carry = (key, policy_params, hyperparams, subs, model_params)
            return carry, log            
            
        # do a batched roll-out from the RDDL model
        def _jax_wrapped_batched_rollout(key, policy_params, hyperparams, 
                                         subs, model_params):
            start = (key, policy_params, hyperparams, subs, model_params)
            steps = jnp.arange(n_steps)
            _, log = jax.lax.scan(_jax_wrapped_batched_step, start, steps)
            log = jax.tree_map(partial(jnp.swapaxes, axis1=0, axis2=1), log)
            return log
        
        return _jax_wrapped_batched_rollout
    
    # ===========================================================================
    # error checks
    # ===========================================================================
    
    def print_jax(self) -> Dict[str, object]:
        '''Returns a dictionary containing the string representations of all 
        Jax compiled expressions from the RDDL file.
        '''
        subs = self.init_values
        params = self.model_params
        key = jax.random.PRNGKey(42)
        printed = {}
        printed['cpfs'] = {name: str(jax.make_jaxpr(expr)(subs, params, key))
                           for (name, expr) in self.cpfs.items()}
        printed['reward'] = str(jax.make_jaxpr(self.reward)(subs, params, key))
        printed['invariants'] = [str(jax.make_jaxpr(expr)(subs, params, key))
                                 for expr in self.invariants]
        printed['preconditions'] = [str(jax.make_jaxpr(expr)(subs, params, key))
                                    for expr in self.preconditions]
        printed['terminations'] = [str(jax.make_jaxpr(expr)(subs, params, key))
                                   for expr in self.termination]
        return printed
        
    @staticmethod
    def _check_valid_op(expr, valid_ops):
        etype, op = expr.etype
        if op not in valid_ops:
            valid_op_str = ','.join(valid_ops.keys())
            raise RDDLNotImplementedError(
                f'{etype} operator {op} is not supported: '
                f'must be in {valid_op_str}.\n' + 
                print_stack_trace(expr))
    
    @staticmethod
    def _check_num_args(expr, required_args):
        actual_args = len(expr.args)
        if actual_args != required_args:
            etype, op = expr.etype
            raise RDDLInvalidNumberOfArgumentsError(
                f'{etype} operator {op} requires {required_args} arguments, '
                f'got {actual_args}.\n' + 
                print_stack_trace(expr))
        
    ERROR_CODES = {
        'NORMAL': 0,
        'INVALID_CAST': 1,
        'INVALID_PARAM_UNIFORM': 2,
        'INVALID_PARAM_NORMAL': 4,
        'INVALID_PARAM_EXPONENTIAL': 8,
        'INVALID_PARAM_WEIBULL': 16,
        'INVALID_PARAM_BERNOULLI': 32,
        'INVALID_PARAM_POISSON': 64,
        'INVALID_PARAM_GAMMA': 128,
        'INVALID_PARAM_BETA': 256,
        'INVALID_PARAM_GEOMETRIC': 512,
        'INVALID_PARAM_PARETO': 1024,
        'INVALID_PARAM_STUDENT': 2048,
        'INVALID_PARAM_GUMBEL': 4096,
        'INVALID_PARAM_LAPLACE': 8192,
        'INVALID_PARAM_CAUCHY': 16384,
        'INVALID_PARAM_GOMPERTZ': 32768,
        'INVALID_PARAM_CHISQUARE': 65536,
        'INVALID_PARAM_KUMARASWAMY': 131072,
        'INVALID_PARAM_DISCRETE': 262144,
        'INVALID_PARAM_KRON_DELTA': 524288,
        'INVALID_PARAM_DIRICHLET': 1048576,
        'INVALID_PARAM_MULTIVARIATE_STUDENT': 2097152,
        'INVALID_PARAM_MULTINOMIAL': 4194304,
        'INVALID_PARAM_BINOMIAL': 8388608,
        'INVALID_PARAM_NEGATIVE_BINOMIAL': 16777216        
    }
    
    INVERSE_ERROR_CODES = {
        0: 'Casting occurred that could result in loss of precision.',
        1: 'Found Uniform(a, b) distribution where a > b.',
        2: 'Found Normal(m, v^2) distribution where v < 0.',
        3: 'Found Exponential(s) distribution where s <= 0.',
        4: 'Found Weibull(k, l) distribution where either k <= 0 or l <= 0.',
        5: 'Found Bernoulli(p) distribution where either p < 0 or p > 1.',
        6: 'Found Poisson(l) distribution where l < 0.',
        7: 'Found Gamma(k, l) distribution where either k <= 0 or l <= 0.',
        8: 'Found Beta(a, b) distribution where either a <= 0 or b <= 0.',
        9: 'Found Geometric(p) distribution where either p < 0 or p > 1.',
        10: 'Found Pareto(k, l) distribution where either k <= 0 or l <= 0.',
        11: 'Found Student(df) distribution where df <= 0.',
        12: 'Found Gumbel(m, s) distribution where s <= 0.',
        13: 'Found Laplace(m, s) distribution where s <= 0.',
        14: 'Found Cauchy(m, s) distribution where s <= 0.',
        15: 'Found Gompertz(k, l) distribution where either k <= 0 or l <= 0.',
        16: 'Found ChiSquare(df) distribution where df <= 0.',
        17: 'Found Kumaraswamy(a, b) distribution where either a <= 0 or b <= 0.',
        18: 'Found Discrete(p) distribution where either p < 0 or p does not sum to 1.',
        19: 'Found KronDelta(x) distribution where x is not int nor bool.',
        20: 'Found Dirichlet(alpha) distribution where alpha < 0.',
        21: 'Found MultivariateStudent(mean, cov, df) distribution where df <= 0.',
        22: 'Found Multinomial(n, p) distribution where either p < 0, p does not sum to 1, or n <= 0.',
        23: 'Found Binomial(n, p) distribution where either p < 0, p > 1, or n <= 0.',
        24: 'Found NegativeBinomial(n, p) distribution where either p < 0, p > 1, or n <= 0.'        
    }
    
    @staticmethod
    def get_error_codes(error):
        '''Given a compacted integer error flag from the execution of Jax, and 
        decomposes it into individual error codes.
        '''
        binary = reversed(bin(error)[2:])
        errors = [i for (i, c) in enumerate(binary) if c == '1']
        return errors
    
    @staticmethod
    def get_error_messages(error):
        '''Given a compacted integer error flag from the execution of Jax, and 
        decomposes it into error strings.
        '''
        codes = JaxRDDLCompiler.get_error_codes(error)
        messages = [JaxRDDLCompiler.INVERSE_ERROR_CODES[i] for i in codes]
        return messages
    
    # ===========================================================================
    # handling of auxiliary data (e.g. model tuning parameters)
    # ===========================================================================
    
    def _unwrap(self, op, expr_id, info):
        sep = JaxRDDLCompiler.MODEL_PARAM_TAG_SEPARATOR
        jax_op, name = op, None
        if isinstance(op, tuple):
            jax_op, param = op
            if param is not None:
                tags, values = param
                if isinstance(tags, tuple):
                    name = sep.join(tags)
                else:
                    name = str(tags)
                name = f'{name}{sep}{expr_id}'
                if name in info:
                    raise Exception(f'Model parameter {name} is already defined.')
                info[name] = values
        return jax_op, name
    
    def get_ids_of_parameterized_expressions(self) -> List[int]:
        '''Returns a list of expression IDs that have tuning parameters.'''
        sep = JaxRDDLCompiler.MODEL_PARAM_TAG_SEPARATOR
        ids = [int(key.split(sep)[-1]) for key in self.model_params] 
        return ids
    
    # ===========================================================================
    # expression compilation
    # ===========================================================================
    
    def _jax(self, expr, info, dtype=None):
        etype, _ = expr.etype
        if etype == 'constant':
            jax_expr = self._jax_constant(expr, info)
        elif etype == 'pvar':
            jax_expr = self._jax_pvar(expr, info)
        elif etype == 'arithmetic':
            jax_expr = self._jax_arithmetic(expr, info)
        elif etype == 'relational':
            jax_expr = self._jax_relational(expr, info)
        elif etype == 'boolean':
            jax_expr = self._jax_logical(expr, info)
        elif etype == 'aggregation':
            jax_expr = self._jax_aggregation(expr, info)
        elif etype == 'func':
            jax_expr = self._jax_functional(expr, info)
        elif etype == 'control':
            jax_expr = self._jax_control(expr, info)
        elif etype == 'randomvar':
            jax_expr = self._jax_random(expr, info)
        elif etype == 'randomvector':
            jax_expr = self._jax_random_vector(expr, info)
        elif etype == 'matrix':
            jax_expr = self._jax_matrix(expr, info)
        else:
            raise RDDLNotImplementedError(
                f'Internal error: expression type {expr} is not supported.\n' + 
                print_stack_trace(expr))
                
        if dtype is not None:
            jax_expr = self._jax_cast(jax_expr, dtype)
        
        return jax_expr
            
    def _jax_cast(self, jax_expr, dtype):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']
        
        def _jax_wrapped_cast(x, params, key):
            val, key, err = jax_expr(x, params, key)
            sample = jnp.asarray(val, dtype=dtype)
            invalid_cast = jnp.logical_and(
                jnp.logical_not(jnp.can_cast(val, dtype)),
                jnp.any(sample != val))
            err |= (invalid_cast * ERR)
            return sample, key, err
        
        return _jax_wrapped_cast
   
    # ===========================================================================
    # leaves
    # ===========================================================================
    
    def _jax_constant(self, expr, info):
        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
        cached_value = self.traced.cached_sim_info(expr)
        
        def _jax_wrapped_constant(x, params, key):
            sample = jnp.asarray(cached_value)
            return sample, key, NORMAL

        return _jax_wrapped_constant
    
    def _jax_pvar_slice(self, _slice):
        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
        
        def _jax_wrapped_pvar_slice(x, params, key):
            return _slice, key, NORMAL
        
        return _jax_wrapped_pvar_slice
            
    def _jax_pvar(self, expr, info):
        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
        var, pvars = expr.args  
        is_value, cached_info = self.traced.cached_sim_info(expr)
        
        # boundary case: free variable is converted to array (0, 1, 2...)
        # boundary case: domain object is converted to canonical integer index
        if is_value:
            cached_value = cached_info

            def _jax_wrapped_object(x, params, key):
                sample = jnp.asarray(cached_value)
                return sample, key, NORMAL
            
            return _jax_wrapped_object
        
        # boundary case: no shape information (e.g. scalar pvar)
        elif cached_info is None:
            
            def _jax_wrapped_pvar_scalar(x, params, key):
                sample = jnp.asarray(x[var])
                return sample, key, NORMAL
            
            return _jax_wrapped_pvar_scalar
        
        # must slice and/or reshape value tensor to match free variables
        else:
            slices, axis, shape, op_code, op_args = cached_info 
        
            # compile nested expressions
            if slices and op_code == RDDLObjectsTracer.NUMPY_OP_CODE.NESTED_SLICE:
                
                jax_nested_expr = [(self._jax(arg, info) 
                                    if _slice is None 
                                    else self._jax_pvar_slice(_slice))
                                   for (arg, _slice) in zip(pvars, slices)]    
                
                def _jax_wrapped_pvar_tensor_nested(x, params, key):
                    error = NORMAL
                    sample = jnp.asarray(x[var])
                    new_slices = [None] * len(jax_nested_expr)
                    for (i, jax_expr) in enumerate(jax_nested_expr):
                        new_slices[i], key, err = jax_expr(x, params, key)
                        error |= err
                    new_slices = tuple(new_slices)
                    sample = sample[new_slices]
                    return sample, key, error
                
                return _jax_wrapped_pvar_tensor_nested
                
            # tensor variable but no nesting  
            else:
    
                def _jax_wrapped_pvar_tensor_non_nested(x, params, key):
                    sample = jnp.asarray(x[var])
                    if slices:
                        sample = sample[slices]
                    if axis:
                        sample = jnp.expand_dims(sample, axis=axis)
                        sample = jnp.broadcast_to(sample, shape=shape)
                    if op_code == RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM:
                        sample = jnp.einsum(sample, *op_args)
                    elif op_code == RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE:
                        sample = jnp.transpose(sample, axes=op_args)
                    return sample, key, NORMAL
                
                return _jax_wrapped_pvar_tensor_non_nested
    
    # ===========================================================================
    # mathematical
    # ===========================================================================
    
    def _jax_unary(self, jax_expr, jax_op, jax_param,
                   at_least_int=False, check_dtype=None):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']

        def _jax_wrapped_unary_op(x, params, key):
            sample, key, err = jax_expr(x, params, key)
            if at_least_int:
                sample = self.ONE * sample
            param = params.get(jax_param, None)
            sample = jax_op(sample, param)
            if check_dtype is not None:
                invalid_cast = jnp.logical_not(jnp.can_cast(sample, check_dtype))
                err |= (invalid_cast * ERR)
            return sample, key, err
        
        return _jax_wrapped_unary_op
    
    def _jax_binary(self, jax_lhs, jax_rhs, jax_op, jax_param,
                    at_least_int=False, check_dtype=None):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']
        
        def _jax_wrapped_binary_op(x, params, key):
            sample1, key, err1 = jax_lhs(x, params, key)
            sample2, key, err2 = jax_rhs(x, params, key)
            if at_least_int:
                sample1 = self.ONE * sample1
                sample2 = self.ONE * sample2
            param = params.get(jax_param, None)
            sample = jax_op(sample1, sample2, param)
            err = err1 | err2
            if check_dtype is not None:
                invalid_cast = jnp.logical_not(jnp.logical_and(
                    jnp.can_cast(sample1, check_dtype),
                    jnp.can_cast(sample2, check_dtype)))
                err |= (invalid_cast * ERR)
            return sample, key, err
        
        return _jax_wrapped_binary_op
    
    def _jax_arithmetic(self, expr, info):
        _, op = expr.etype
        valid_ops = self.ARITHMETIC_OPS
        JaxRDDLCompiler._check_valid_op(expr, valid_ops)
                    
        args = expr.args
        n = len(args)
        
        if n == 1 and op == '-':
            arg, = args
            jax_expr = self._jax(arg, info)
            jax_op, jax_param = self._unwrap(self.NEGATIVE, expr.id, info)
            return self._jax_unary(jax_expr, jax_op, jax_param, at_least_int=True)
                    
        elif n == 2:
            lhs, rhs = args
            jax_lhs = self._jax(lhs, info)
            jax_rhs = self._jax(rhs, info)
            jax_op, jax_param = self._unwrap(valid_ops[op], expr.id, info)
            return self._jax_binary(
                jax_lhs, jax_rhs, jax_op, jax_param, at_least_int=True)
        
        JaxRDDLCompiler._check_num_args(expr, 2)
    
    def _jax_relational(self, expr, info):
        _, op = expr.etype
        valid_ops = self.RELATIONAL_OPS
        JaxRDDLCompiler._check_valid_op(expr, valid_ops)
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        lhs, rhs = expr.args
        jax_lhs = self._jax(lhs, info)
        jax_rhs = self._jax(rhs, info)
        jax_op, jax_param = self._unwrap(valid_ops[op], expr.id, info)
        return self._jax_binary(
            jax_lhs, jax_rhs, jax_op, jax_param, at_least_int=True)
           
    def _jax_logical(self, expr, info):
        _, op = expr.etype
        valid_ops = self.LOGICAL_OPS    
        JaxRDDLCompiler._check_valid_op(expr, valid_ops)
                
        args = expr.args
        n = len(args)
        
        if n == 1 and op == '~':
            arg, = args
            jax_expr = self._jax(arg, info)
            jax_op, jax_param = self._unwrap(self.LOGICAL_NOT, expr.id, info)
            return self._jax_unary(jax_expr, jax_op, jax_param, check_dtype=bool)
        
        elif n == 2:
            lhs, rhs = args
            jax_lhs = self._jax(lhs, info)
            jax_rhs = self._jax(rhs, info)
            jax_op, jax_param = self._unwrap(valid_ops[op], expr.id, info)
            return self._jax_binary(
                jax_lhs, jax_rhs, jax_op, jax_param, check_dtype=bool)
        
        JaxRDDLCompiler._check_num_args(expr, 2)
    
    def _jax_aggregation(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']
        
        _, op = expr.etype
        valid_ops = self.AGGREGATION_OPS      
        JaxRDDLCompiler._check_valid_op(expr, valid_ops) 
        is_floating = op not in self.AGGREGATION_BOOL
        
        * _, arg = expr.args  
        _, axes = self.traced.cached_sim_info(expr)
        
        jax_expr = self._jax(arg, info)
        jax_op, jax_param = self._unwrap(valid_ops[op], expr.id, info)
        
        def _jax_wrapped_aggregation(x, params, key):
            sample, key, err = jax_expr(x, params, key)
            if is_floating:
                sample = self.ONE * sample
            else:
                invalid_cast = jnp.logical_not(jnp.can_cast(sample, bool))
                err |= (invalid_cast * ERR)
            param = params.get(jax_param, None)
            sample = jax_op(sample, axis=axes, param=param)
            return sample, key, err
        
        return _jax_wrapped_aggregation
               
    def _jax_functional(self, expr, info):
        _, op = expr.etype
        
        # unary function
        if op in self.KNOWN_UNARY:
            JaxRDDLCompiler._check_num_args(expr, 1)                            
            arg, = expr.args
            jax_expr = self._jax(arg, info)
            jax_op, jax_param = self._unwrap(self.KNOWN_UNARY[op], expr.id, info)
            return self._jax_unary(jax_expr, jax_op, jax_param, at_least_int=True)
            
        # binary function
        elif op in self.KNOWN_BINARY:
            JaxRDDLCompiler._check_num_args(expr, 2)                
            lhs, rhs = expr.args
            jax_lhs = self._jax(lhs, info)
            jax_rhs = self._jax(rhs, info)
            jax_op, jax_param = self._unwrap(self.KNOWN_BINARY[op], expr.id, info)
            return self._jax_binary(
                jax_lhs, jax_rhs, jax_op, jax_param, at_least_int=True)
        
        raise RDDLNotImplementedError(
            f'Function {op} is not supported.\n' + 
            print_stack_trace(expr))   
    
    # ===========================================================================
    # control flow
    # ===========================================================================
    
    def _jax_control(self, expr, info):
        _, op = expr.etype        
        if op == 'if':
            return self._jax_if(expr, info)
        elif op == 'switch':
            return self._jax_switch(expr, info)
        
        raise RDDLNotImplementedError(
            f'Control operator {op} is not supported.\n' + 
            print_stack_trace(expr))   
    
    def _jax_if_helper(self):
        
        def _jax_wrapped_if_calc_exact(c, a, b, param):
            return jnp.where(c, a, b)
        
        return _jax_wrapped_if_calc_exact
    
    def _jax_if(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_CAST']
        JaxRDDLCompiler._check_num_args(expr, 3)
        jax_if, jax_param = self._unwrap(self._jax_if_helper(), expr.id, info)
        
        pred, if_true, if_false = expr.args        
        jax_pred = self._jax(pred, info)
        jax_true = self._jax(if_true, info)
        jax_false = self._jax(if_false, info)
        
        def _jax_wrapped_if_then_else(x, params, key):
            sample1, key, err1 = jax_pred(x, params, key)
            sample2, key, err2 = jax_true(x, params, key)
            sample3, key, err3 = jax_false(x, params, key)
            param = params.get(jax_param, None)
            sample = jax_if(sample1, sample2, sample3, param)
            err = err1 | err2 | err3
            invalid_cast = jnp.logical_not(jnp.can_cast(sample1, bool))
            err |= (invalid_cast * ERR)
            return sample, key, err
            
        return _jax_wrapped_if_then_else
    
    def _jax_switch_helper(self):
        
        def _jax_wrapped_switch_calc_exact(pred, cases, param):
            pred = pred[jnp.newaxis, ...]
            sample = jnp.take_along_axis(cases, pred, axis=0)
            assert sample.shape[0] == 1
            return sample[0, ...]

        return _jax_wrapped_switch_calc_exact
        
    def _jax_switch(self, expr, info):
        pred, *_ = expr.args             
        jax_pred = self._jax(pred, info)
        jax_switch, jax_param = self._unwrap(
            self._jax_switch_helper(), expr.id, info)
        
        # wrap cases as JAX expressions
        cases, default = self.traced.cached_sim_info(expr) 
        jax_default = None if default is None else self._jax(default, info)
        jax_cases = [(jax_default if _case is None else self._jax(_case, info))
                     for _case in cases]
                    
        def _jax_wrapped_switch(x, params, key):
            
            # sample predicate
            sample_pred, key, err = jax_pred(x, params, key) 
            
            # sample cases
            sample_cases = [None] * len(jax_cases)
            for (i, jax_case) in enumerate(jax_cases):
                sample_cases[i], key, err_case = jax_case(x, params, key)
                err |= err_case                
            sample_cases = jnp.asarray(sample_cases)
            
            # predicate (enum) is an integer - use it to extract from case array
            param = params.get(jax_param, None)
            sample = jax_switch(sample_pred, sample_cases, param)
            return sample, key, err    
        
        return _jax_wrapped_switch
    
    # ===========================================================================
    # random variables
    # ===========================================================================
    
    # distributions with complete reparameterization support:
    # KronDelta: complete
    # DiracDelta: complete
    # Uniform: complete
    # Bernoulli: complete (subclass uses Gumbel-softmax)
    # Normal: complete
    # Exponential: complete
    # Weibull: complete
    # Pareto: complete
    # Gumbel: complete
    # Laplace: complete
    # Cauchy: complete
    # Gompertz: complete
    # Kumaraswamy: complete
    # Discrete: complete (subclass uses Gumbel-softmax)
    # UnnormDiscrete: complete (subclass uses Gumbel-softmax)
    # Discrete(p): complete (subclass uses Gumbel-softmax)
    # UnnormDiscrete(p): complete (subclass uses Gumbel-softmax)
    
    # distributions with incomplete reparameterization support (TODO):
    # Binomial: (use truncation and Gumbel-softmax)
    # NegativeBinomial: (no reparameterization)
    # Poisson: (use truncation and Gumbel-softmax)
    # Gamma, ChiSquare: (no shape reparameterization)
    # Beta: (no reparameterization)
    # Geometric: (implement safe floor)
    # Student: (no reparameterization)
    
    def _jax_random(self, expr, info):
        _, name = expr.etype
        if name == 'KronDelta':
            return self._jax_kron(expr, info)        
        elif name == 'DiracDelta':
            return self._jax_dirac(expr, info)
        elif name == 'Uniform':
            return self._jax_uniform(expr, info)
        elif name == 'Bernoulli':
            return self._jax_bernoulli(expr, info)
        elif name == 'Normal':
            return self._jax_normal(expr, info)  
        elif name == 'Poisson':
            return self._jax_poisson(expr, info)
        elif name == 'Exponential':
            return self._jax_exponential(expr, info)
        elif name == 'Weibull':
            return self._jax_weibull(expr, info) 
        elif name == 'Gamma':
            return self._jax_gamma(expr, info)
        elif name == 'Binomial':
            return self._jax_binomial(expr, info)
        elif name == 'NegativeBinomial':
            return self._jax_negative_binomial(expr, info)
        elif name == 'Beta':
            return self._jax_beta(expr, info)
        elif name == 'Geometric':
            return self._jax_geometric(expr, info)
        elif name == 'Pareto':
            return self._jax_pareto(expr, info)
        elif name == 'Student':
            return self._jax_student(expr, info)
        elif name == 'Gumbel':
            return self._jax_gumbel(expr, info)
        elif name == 'Laplace':
            return self._jax_laplace(expr, info)
        elif name == 'Cauchy':
            return self._jax_cauchy(expr, info)
        elif name == 'Gompertz':
            return self._jax_gompertz(expr, info)
        elif name == 'ChiSquare':
            return self._jax_chisquare(expr, info)
        elif name == 'Kumaraswamy':
            return self._jax_kumaraswamy(expr, info)
        elif name == 'Discrete':
            return self._jax_discrete(expr, info, unnorm=False)
        elif name == 'UnnormDiscrete':
            return self._jax_discrete(expr, info, unnorm=True)
        elif name == 'Discrete(p)':
            return self._jax_discrete_pvar(expr, info, unnorm=False)
        elif name == 'UnnormDiscrete(p)':
            return self._jax_discrete_pvar(expr, info, unnorm=True)
        else:
            raise RDDLNotImplementedError(
                f'Distribution {name} is not supported.\n' + 
                print_stack_trace(expr))
        
    def _jax_kron(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_KRON_DELTA']
        JaxRDDLCompiler._check_num_args(expr, 1)
        arg, = expr.args
        arg = self._jax(arg, info)
        
        # just check that the sample can be cast to int
        def _jax_wrapped_distribution_kron(x, params, key):
            sample, key, err = arg(x, params, key)
            invalid_cast = jnp.logical_not(jnp.can_cast(sample, self.INT))
            err |= (invalid_cast * ERR)
            return sample, key, err
                        
        return _jax_wrapped_distribution_kron
    
    def _jax_dirac(self, expr, info):
        JaxRDDLCompiler._check_num_args(expr, 1)
        arg, = expr.args
        arg = self._jax(arg, info, dtype=self.REAL)
        return arg
    
    def _jax_uniform(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_UNIFORM']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_lb, arg_ub = expr.args
        jax_lb = self._jax(arg_lb, info)
        jax_ub = self._jax(arg_ub, info)
        
        # reparameterization trick U(a, b) = a + (b - a) * U(0, 1)
        def _jax_wrapped_distribution_uniform(x, params, key):
            lb, key, err1 = jax_lb(x, params, key)
            ub, key, err2 = jax_ub(x, params, key)
            key, subkey = random.split(key)
            U = random.uniform(key=subkey, shape=jnp.shape(lb), dtype=self.REAL)
            sample = lb + (ub - lb) * U
            out_of_bounds = jnp.logical_not(jnp.all(lb <= ub))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_uniform
    
    def _jax_normal(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_NORMAL']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_mean, arg_var = expr.args
        jax_mean = self._jax(arg_mean, info)
        jax_var = self._jax(arg_var, info)
        
        # reparameterization trick N(m, s^2) = m + s * N(0, 1)
        def _jax_wrapped_distribution_normal(x, params, key):
            mean, key, err1 = jax_mean(x, params, key)
            var, key, err2 = jax_var(x, params, key)
            std = jnp.sqrt(var)
            key, subkey = random.split(key)
            Z = random.normal(key=subkey, shape=jnp.shape(mean), dtype=self.REAL)
            sample = mean + std * Z
            out_of_bounds = jnp.logical_not(jnp.all(var >= 0))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_normal
    
    def _jax_exponential(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_EXPONENTIAL']
        JaxRDDLCompiler._check_num_args(expr, 1)
        
        arg_scale, = expr.args
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick Exp(s) = s * Exp(1)
        def _jax_wrapped_distribution_exp(x, params, key):
            scale, key, err = jax_scale(x, params, key)
            key, subkey = random.split(key)
            Exp1 = random.exponential(
                key=subkey, shape=jnp.shape(scale), dtype=self.REAL)
            sample = scale * Exp1
            out_of_bounds = jnp.logical_not(jnp.all(scale > 0))
            err |= (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_exp
    
    def _jax_weibull(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_WEIBULL']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_shape, arg_scale = expr.args
        jax_shape = self._jax(arg_shape, info)
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick W(s, r) = r * (-ln(1 - U(0, 1))) ** (1 / s)
        def _jax_wrapped_distribution_weibull(x, params, key):
            shape, key, err1 = jax_shape(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            U = random.uniform(key=subkey, shape=jnp.shape(scale), dtype=self.REAL)
            sample = scale * jnp.power(-jnp.log1p(-U), 1.0 / shape)
            out_of_bounds = jnp.logical_not(jnp.all((shape > 0) & (scale > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_weibull
    
    def _jax_bernoulli_helper(self):
        
        def _jax_wrapped_calc_bernoulli_exact(key, prob, param):
            return random.bernoulli(key, prob)
        
        return _jax_wrapped_calc_bernoulli_exact
        
    def _jax_bernoulli(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BERNOULLI']
        JaxRDDLCompiler._check_num_args(expr, 1)
        jax_bern, jax_param = self._unwrap(
            self._jax_bernoulli_helper(), expr.id, info)
        
        arg_prob, = expr.args
        jax_prob = self._jax(arg_prob, info)
        
        # uses the implicit JAX subroutine
        def _jax_wrapped_distribution_bernoulli(x, params, key):
            prob, key, err = jax_prob(x, params, key)
            key, subkey = random.split(key)
            param = params.get(jax_param, None)
            sample = jax_bern(subkey, prob, param)
            out_of_bounds = jnp.logical_not(jnp.all((prob >= 0) & (prob <= 1)))
            err |= (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_bernoulli
    
    def _jax_poisson(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_POISSON']
        JaxRDDLCompiler._check_num_args(expr, 1)
        
        arg_rate, = expr.args
        jax_rate = self._jax(arg_rate, info)
        
        # uses the implicit JAX subroutine
        def _jax_wrapped_distribution_poisson(x, params, key):
            rate, key, err = jax_rate(x, params, key)
            key, subkey = random.split(key)
            sample = random.poisson(key=subkey, lam=rate, dtype=self.INT)
            out_of_bounds = jnp.logical_not(jnp.all(rate >= 0))
            err |= (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_poisson
    
    def _jax_gamma(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_GAMMA']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_shape, arg_scale = expr.args
        jax_shape = self._jax(arg_shape, info)
        jax_scale = self._jax(arg_scale, info)
        
        # partial reparameterization trick Gamma(s, r) = r * Gamma(s, 1)
        # uses the implicit JAX subroutine for Gamma(s, 1) 
        def _jax_wrapped_distribution_gamma(x, params, key):
            shape, key, err1 = jax_shape(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            Gamma = random.gamma(key=subkey, a=shape, dtype=self.REAL)
            sample = scale * Gamma
            out_of_bounds = jnp.logical_not(jnp.all((shape > 0) & (scale > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_gamma
    
    def _jax_binomial(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BINOMIAL']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_trials, arg_prob = expr.args
        jax_trials = self._jax(arg_trials, info)
        jax_prob = self._jax(arg_prob, info)
        
        # uses the JAX substrate of tensorflow-probability
        def _jax_wrapped_distribution_binomial(x, params, key):
            trials, key, err2 = jax_trials(x, params, key)       
            prob, key, err1 = jax_prob(x, params, key)
            trials = jnp.asarray(trials, self.REAL)
            prob = jnp.asarray(prob, self.REAL)
            key, subkey = random.split(key)
            dist = tfp.distributions.Binomial(total_count=trials, probs=prob)
            sample = dist.sample(seed=subkey).astype(self.INT)
            out_of_bounds = jnp.logical_not(jnp.all(
                (prob >= 0) & (prob <= 1) & (trials >= 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_binomial
    
    def _jax_negative_binomial(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_NEGATIVE_BINOMIAL']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_trials, arg_prob = expr.args
        jax_trials = self._jax(arg_trials, info)
        jax_prob = self._jax(arg_prob, info)
        
        # uses the JAX substrate of tensorflow-probability
        def _jax_wrapped_distribution_negative_binomial(x, params, key):
            trials, key, err2 = jax_trials(x, params, key)       
            prob, key, err1 = jax_prob(x, params, key)
            trials = jnp.asarray(trials, self.REAL)
            prob = jnp.asarray(prob, self.REAL)
            key, subkey = random.split(key)
            dist = tfp.distributions.NegativeBinomial(
                total_count=trials, probs=prob)
            sample = dist.sample(seed=subkey).astype(self.INT)
            out_of_bounds = jnp.logical_not(jnp.all(
                (prob >= 0) & (prob <= 1) & (trials > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_negative_binomial    
        
    def _jax_beta(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_BETA']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_shape, arg_rate = expr.args
        jax_shape = self._jax(arg_shape, info)
        jax_rate = self._jax(arg_rate, info)
        
        # uses the implicit JAX subroutine
        def _jax_wrapped_distribution_beta(x, params, key):
            shape, key, err1 = jax_shape(x, params, key)
            rate, key, err2 = jax_rate(x, params, key)
            key, subkey = random.split(key)
            sample = random.beta(key=subkey, a=shape, b=rate)
            out_of_bounds = jnp.logical_not(jnp.all((shape > 0) & (rate > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_beta
    
    def _jax_geometric(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_GEOMETRIC']
        JaxRDDLCompiler._check_num_args(expr, 1)
        
        arg_prob, = expr.args
        jax_prob = self._jax(arg_prob, info)
        floor_op, jax_param = self._unwrap(
            self.KNOWN_UNARY['floor'], expr.id, info)
        
        # reparameterization trick Geom(p) = floor(ln(U(0, 1)) / ln(p)) + 1
        def _jax_wrapped_distribution_geometric(x, params, key):
            prob, key, err = jax_prob(x, params, key)
            key, subkey = random.split(key)
            U = random.uniform(key=subkey, shape=jnp.shape(prob), dtype=self.REAL)
            param = params.get(jax_param, None)
            sample = floor_op(jnp.log1p(-U) / jnp.log1p(-prob), param) + 1
            out_of_bounds = jnp.logical_not(jnp.all((prob >= 0) & (prob <= 1)))
            err |= (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_geometric
    
    def _jax_pareto(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_PARETO']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_shape, arg_scale = expr.args
        jax_shape = self._jax(arg_shape, info)
        jax_scale = self._jax(arg_scale, info)
        
        # partial reparameterization trick Pareto(s, r) = r * Pareto(s, 1)
        # uses the implicit JAX subroutine for Pareto(s, 1) 
        def _jax_wrapped_distribution_pareto(x, params, key):
            shape, key, err1 = jax_shape(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            sample = scale * random.pareto(key=subkey, b=shape)
            out_of_bounds = jnp.logical_not(jnp.all((shape > 0) & (scale > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_pareto
    
    def _jax_student(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_STUDENT']
        JaxRDDLCompiler._check_num_args(expr, 1)
        
        arg_df, = expr.args
        jax_df = self._jax(arg_df, info)
        
        # uses the implicit JAX subroutine for student(df)
        def _jax_wrapped_distribution_t(x, params, key):
            df, key, err = jax_df(x, params, key)
            key, subkey = random.split(key)
            sample = random.t(key=subkey, df=df, shape=jnp.shape(df))
            out_of_bounds = jnp.logical_not(jnp.all(df > 0))
            err |= (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_t
    
    def _jax_gumbel(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_GUMBEL']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_mean, arg_scale = expr.args
        jax_mean = self._jax(arg_mean, info)
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick Gumbel(m, s) = m + s * Gumbel(0, 1)
        def _jax_wrapped_distribution_gumbel(x, params, key):
            mean, key, err1 = jax_mean(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            Gumbel01 = random.gumbel(
                key=subkey, shape=jnp.shape(mean), dtype=self.REAL)
            sample = mean + scale * Gumbel01
            out_of_bounds = jnp.logical_not(jnp.all(scale > 0))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_gumbel
    
    def _jax_laplace(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_LAPLACE']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_mean, arg_scale = expr.args
        jax_mean = self._jax(arg_mean, info)
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick Laplace(m, s) = m + s * Laplace(0, 1)
        def _jax_wrapped_distribution_laplace(x, params, key):
            mean, key, err1 = jax_mean(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            Laplace01 = random.laplace(
                key=subkey, shape=jnp.shape(mean), dtype=self.REAL)
            sample = mean + scale * Laplace01
            out_of_bounds = jnp.logical_not(jnp.all(scale > 0))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_laplace
    
    def _jax_cauchy(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_CAUCHY']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_mean, arg_scale = expr.args
        jax_mean = self._jax(arg_mean, info)
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick Cauchy(m, s) = m + s * Cauchy(0, 1)
        def _jax_wrapped_distribution_cauchy(x, params, key):
            mean, key, err1 = jax_mean(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            Cauchy01 = random.cauchy(
                key=subkey, shape=jnp.shape(mean), dtype=self.REAL)
            sample = mean + scale * Cauchy01
            out_of_bounds = jnp.logical_not(jnp.all(scale > 0))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_cauchy
    
    def _jax_gompertz(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_GOMPERTZ']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_shape, arg_scale = expr.args
        jax_shape = self._jax(arg_shape, info)
        jax_scale = self._jax(arg_scale, info)
        
        # reparameterization trick Gompertz(s, r) = ln(1 - log(U(0, 1)) / s) / r
        def _jax_wrapped_distribution_gompertz(x, params, key):
            shape, key, err1 = jax_shape(x, params, key)
            scale, key, err2 = jax_scale(x, params, key)
            key, subkey = random.split(key)
            U = random.uniform(key=subkey, shape=jnp.shape(scale), dtype=self.REAL)
            sample = jnp.log(1.0 - jnp.log1p(-U) / shape) / scale
            out_of_bounds = jnp.logical_not(jnp.all((shape > 0) & (scale > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_gompertz
    
    def _jax_chisquare(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_CHISQUARE']
        JaxRDDLCompiler._check_num_args(expr, 1)
        
        arg_df, = expr.args
        jax_df = self._jax(arg_df, info)
        
        # use the fact that ChiSquare(df) = Gamma(df/2, 2)
        def _jax_wrapped_distribution_chisquare(x, params, key):
            df, key, err1 = jax_df(x, params, key)
            key, subkey = random.split(key)
            shape = df / 2.0
            Gamma = random.gamma(key=subkey, a=shape, dtype=self.REAL)
            sample = 2.0 * Gamma
            out_of_bounds = jnp.logical_not(jnp.all(df > 0))
            err = err1 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_chisquare
    
    def _jax_kumaraswamy(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_KUMARASWAMY']
        JaxRDDLCompiler._check_num_args(expr, 2)
        
        arg_a, arg_b = expr.args
        jax_a = self._jax(arg_a, info)
        jax_b = self._jax(arg_b, info)
        
        # uses the reparameterization K(a, b) = (1 - (1 - U(0, 1))^{1/b})^{1/a}
        def _jax_wrapped_distribution_kumaraswamy(x, params, key):
            a, key, err1 = jax_a(x, params, key)
            b, key, err2 = jax_b(x, params, key)
            key, subkey = random.split(key)
            U = random.uniform(key=subkey, shape=jnp.shape(a), dtype=self.REAL)            
            sample = jnp.power(1.0 - jnp.power(U, 1.0 / b), 1.0 / a)
            out_of_bounds = jnp.logical_not(jnp.all((a > 0) & (b > 0)))
            err = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, err
        
        return _jax_wrapped_distribution_kumaraswamy
    
    # ===========================================================================
    # random variables with enum support
    # ===========================================================================
    
    def _jax_discrete_helper(self):
        
        def _jax_wrapped_discrete_calc_exact(key, prob, param):
            logits = jnp.log(prob)
            sample = random.categorical(key=key, logits=logits, axis=-1)
            out_of_bounds = jnp.logical_not(jnp.logical_and(
                jnp.all(prob >= 0),
                jnp.allclose(jnp.sum(prob, axis=-1), 1.0)))
            return sample, out_of_bounds
        
        return _jax_wrapped_discrete_calc_exact
            
    def _jax_discrete(self, expr, info, unnorm):
        NORMAL = JaxRDDLCompiler.ERROR_CODES['NORMAL']
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_DISCRETE']
        jax_discrete, jax_param = self._unwrap(
            self._jax_discrete_helper(), expr.id, info)
        
        ordered_args = self.traced.cached_sim_info(expr)
        jax_probs = [self._jax(arg, info) for arg in ordered_args]
        
        def _jax_wrapped_distribution_discrete(x, params, key):
            
            # sample case probabilities and normalize as needed
            error = NORMAL
            prob = [None] * len(jax_probs)
            for (i, jax_prob) in enumerate(jax_probs):
                prob[i], key, error_pdf = jax_prob(x, params, key)
                error |= error_pdf
            prob = jnp.stack(prob, axis=-1)
            if unnorm:
                normalizer = jnp.sum(prob, axis=-1, keepdims=True)
                prob = prob / normalizer
            
            # dispatch to sampling subroutine
            key, subkey = random.split(key)
            param = params.get(jax_param, None)
            sample, out_of_bounds = jax_discrete(subkey, prob, param)
            error |= (out_of_bounds * ERR)
            return sample, key, error
        
        return _jax_wrapped_distribution_discrete
    
    def _jax_discrete_pvar(self, expr, info, unnorm):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_DISCRETE']
        JaxRDDLCompiler._check_num_args(expr, 1)
        jax_discrete, jax_param = self._unwrap(
            self._jax_discrete_helper(), expr.id, info)
        
        _, args = expr.args
        arg, = args
        jax_probs = self._jax(arg, info)

        def _jax_wrapped_distribution_discrete_pvar(x, params, key):
            
            # sample probabilities
            prob, key, error = jax_probs(x, params, key)
            if unnorm:
                normalizer = jnp.sum(prob, axis=-1, keepdims=True)
                prob = prob / normalizer
            
            # dispatch to sampling subroutine
            key, subkey = random.split(key)
            param = params.get(jax_param, None)
            sample, out_of_bounds = jax_discrete(subkey, prob, param)
            error |= (out_of_bounds * ERR)
            return sample, key, error
        
        return _jax_wrapped_distribution_discrete_pvar

    # ===========================================================================
    # random vectors
    # ===========================================================================
    
    def _jax_random_vector(self, expr, info):
        _, name = expr.etype
        if name == 'MultivariateNormal':
            return self._jax_multivariate_normal(expr, info)   
        elif name == 'MultivariateStudent':
            return self._jax_multivariate_student(expr, info)  
        elif name == 'Dirichlet':
            return self._jax_dirichlet(expr, info)
        elif name == 'Multinomial':
            return self._jax_multinomial(expr, info)
        else:
            raise RDDLNotImplementedError(
                f'Distribution {name} is not supported.\n' + 
                print_stack_trace(expr))
    
    def _jax_multivariate_normal(self, expr, info): 
        _, args = expr.args
        mean, cov = args
        jax_mean = self._jax(mean, info)
        jax_cov = self._jax(cov, info)
        index, = self.traced.cached_sim_info(expr)
        
        # reparameterization trick MN(m, LL') = LZ + m, where Z ~ Normal(0, 1)
        def _jax_wrapped_distribution_multivariate_normal(x, params, key):
            
            # sample the mean and covariance
            sample_mean, key, err1 = jax_mean(x, params, key)
            sample_cov, key, err2 = jax_cov(x, params, key)
            
            # sample Normal(0, 1)
            key, subkey = random.split(key)
            Z = random.normal(
                key=subkey,
                shape=jnp.shape(sample_mean) + (1,),
                dtype=self.REAL)       
            
            # compute L s.t. cov = L * L' and reparameterize
            L = jnp.linalg.cholesky(sample_cov)
            sample = jnp.matmul(L, Z)[..., 0] + sample_mean
            sample = jnp.moveaxis(sample, source=-1, destination=index)
            err = err1 | err2
            return sample, key, err
        
        return _jax_wrapped_distribution_multivariate_normal
    
    def _jax_multivariate_student(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_MULTIVARIATE_STUDENT']
        
        _, args = expr.args
        mean, cov, df = args
        jax_mean = self._jax(mean, info)
        jax_cov = self._jax(cov, info)
        jax_df = self._jax(df, info)
        index, = self.traced.cached_sim_info(expr)
        
        # reparameterization trick MN(m, LL') = LZ + m, where Z ~ StudentT(0, 1)
        def _jax_wrapped_distribution_multivariate_student(x, params, key):
            
            # sample the mean and covariance and degrees of freedom
            sample_mean, key, err1 = jax_mean(x, params, key)
            sample_cov, key, err2 = jax_cov(x, params, key)
            sample_df, key, err3 = jax_df(x, params, key)
            out_of_bounds = jnp.logical_not(jnp.all(sample_df > 0))
            
            # sample StudentT(0, 1, df) -- broadcast df to same shape as cov
            sample_df = sample_df[..., jnp.newaxis, jnp.newaxis]
            sample_df = jnp.broadcast_to(sample_df, shape=sample_mean.shape + (1,))
            key, subkey = random.split(key)
            Z = random.t(
                key=subkey, 
                df=sample_df, 
                shape=jnp.shape(sample_df),
                dtype=self.REAL)   
            
            # compute L s.t. cov = L * L' and reparameterize
            L = jnp.linalg.cholesky(sample_cov)
            sample = jnp.matmul(L, Z)[..., 0] + sample_mean
            sample = jnp.moveaxis(sample, source=-1, destination=index)
            error = err1 | err2 | err3 | (out_of_bounds * ERR)
            return sample, key, error
        
        return _jax_wrapped_distribution_multivariate_student
    
    def _jax_dirichlet(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_DIRICHLET']
        
        _, args = expr.args
        alpha, = args
        jax_alpha = self._jax(alpha, info)
        index, = self.traced.cached_sim_info(expr)
        
        # sample Gamma(alpha_i, 1) and normalize across i
        def _jax_wrapped_distribution_dirichlet(x, params, key):
            alpha, key, error = jax_alpha(x, params, key)
            out_of_bounds = jnp.logical_not(jnp.all(alpha > 0))
            error |= (out_of_bounds * ERR)
            key, subkey = random.split(key)
            Gamma = random.gamma(key=subkey, a=alpha)
            sample = Gamma / jnp.sum(Gamma, axis=-1, keepdims=True)
            sample = jnp.moveaxis(sample, source=-1, destination=index)
            return sample, key, error
        
        return _jax_wrapped_distribution_dirichlet
    
    def _jax_multinomial(self, expr, info):
        ERR = JaxRDDLCompiler.ERROR_CODES['INVALID_PARAM_MULTINOMIAL']
        
        _, args = expr.args
        trials, prob = args
        jax_trials = self._jax(trials, info)
        jax_prob = self._jax(prob, info)
        index, = self.traced.cached_sim_info(expr)
        
        def _jax_wrapped_distribution_multinomial(x, params, key):
            trials, key, err1 = jax_trials(x, params, key)
            prob, key, err2 = jax_prob(x, params, key)
            trials = jnp.asarray(trials, self.REAL)
            prob = jnp.asarray(prob, self.REAL)
            key, subkey = random.split(key)
            dist = tfp.distributions.Multinomial(total_count=trials, probs=prob)
            sample = dist.sample(seed=subkey).astype(self.INT)
            sample = jnp.moveaxis(sample, source=-1, destination=index)
            out_of_bounds = jnp.logical_not(jnp.all(
                (prob >= 0)
                & jnp.allclose(jnp.sum(prob, axis=-1), 1.0)
                & (trials >= 0)))
            error = err1 | err2 | (out_of_bounds * ERR)
            return sample, key, error            
        
        return _jax_wrapped_distribution_multinomial
    
    # ===========================================================================
    # matrix algebra
    # ===========================================================================
    
    def _jax_matrix(self, expr, info):
        _, op = expr.etype
        if op == 'det':
            return self._jax_matrix_det(expr, info)
        elif op == 'inverse':
            return self._jax_matrix_inv(expr, info, pseudo=False)
        elif op == 'pinverse':
            return self._jax_matrix_inv(expr, info, pseudo=True)
        elif op == 'cholesky':
            return self._jax_matrix_cholesky(expr, info)
        else:
            raise RDDLNotImplementedError(
                f'Matrix operation {op} is not supported.\n' + 
                print_stack_trace(expr))
    
    def _jax_matrix_det(self, expr, info):
        * _, arg = expr.args
        jax_arg = self._jax(arg, info)
        
        def _jax_wrapped_matrix_operation_det(x, params, key):
            sample_arg, key, error = jax_arg(x, params, key)
            sample = jnp.linalg.det(sample_arg)
            return sample, key, error
        
        return _jax_wrapped_matrix_operation_det
    
    def _jax_matrix_inv(self, expr, info, pseudo):
        _, arg = expr.args
        jax_arg = self._jax(arg, info)
        indices = self.traced.cached_sim_info(expr)
        op = jnp.linalg.pinv if pseudo else jnp.linalg.inv
        
        def _jax_wrapped_matrix_operation_inv(x, params, key):
            sample_arg, key, error = jax_arg(x, params, key)
            sample = op(sample_arg)
            sample = jnp.moveaxis(sample, source=(-2, -1), destination=indices)
            return sample, key, error
        
        return _jax_wrapped_matrix_operation_inv
    
    def _jax_matrix_cholesky(self, expr, info):
        _, arg = expr.args
        jax_arg = self._jax(arg, info)
        indices = self.traced.cached_sim_info(expr)
        op = jnp.linalg.cholesky
        
        def _jax_wrapped_matrix_operation_cholesky(x, params, key):
            sample_arg, key, error = jax_arg(x, params, key)
            sample = op(sample_arg)
            sample = jnp.moveaxis(sample, source=(-2, -1), destination=indices)
            return sample, key, error
        
        return _jax_wrapped_matrix_operation_cholesky
            



from pathlib import Path
from typing import Optional
from pyRDDLGym.Core.Grounder.RDDLGrounder import RDDLGrounder
from pyRDDLGym.XADD.RDDLModelXADD import RDDLModelWXADD
from pyRDDLGym.Core.Parser.RDDLReader import RDDLReader
from pyRDDLGym.Core.Parser.parser import RDDLParser
from pyRDDLGym.Examples.ExampleManager import ExampleManager


def test_xadd(
        env_name: str = 'wildfire',
        cpf: Optional[str] = None,
        save_graph: bool = False,
        simulation: bool = False,
):
    env_info = ExampleManager.GetEnvInfo(env_name)
    domain = env_info.get_domain()
    instance = env_info.get_instance(0)
    
    # Read and parse domain and instance
    reader = RDDLReader(domain, instance)
    domain = reader.rddltxt
    parser = RDDLParser(None, False)
    parser.build()

    # Parse RDDL file
    rddl_ast = parser.parse(domain)

    # Ground domain
    grounder = RDDLGrounder(rddl_ast)
    model = grounder.Ground()

    # XADD compilation
    xadd_model = RDDLModelWXADD(model, simulation=simulation)
    xadd_model.compile()
    context = xadd_model._context
    
    if save_graph:
        Path(f'tmp/{env_name}').mkdir(parents=True, exist_ok=True)
        
    if cpf is not None or cpf == 'reward':
        if cpf == 'reward':
            expr = xadd_model.reward
        else:
            expr = xadd_model.cpfs.get(f"{cpf}'")
        if expr is None:
            raise AttributeError(f"Cannot retrieve {cpf}' from 'model.cpfs'")
        print(f"cpf {cpf}':", end='\n')
        xadd_model.print(expr)
        if save_graph:
            f_path = f"{env_name}/{env_name}_inst0_{cpf}"
            context.save_graph(expr, f_path)
    else:
        for cpf_, expr in xadd_model.cpfs.items():
            print(f"cpf {cpf_}:", end='\n')
            xadd_model.print(expr)
            if save_graph:
                cpf = cpf_.strip("'")
                f_path = f"{env_name}/{env_name}_inst0_{cpf}"
                context.save_graph(expr, f_path)
    # xadd_model.print(xadd_model.reward)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument('--env', type=str, default='wildfire',
                        help='The name of the RDDL environment')
    parser.add_argument('--cpf', type=str, default=None,
                        help='If specified, only print out this CPF')
    parser.add_argument('--save_graph', action='store_true',
                        help='Save the graph as pdf file')
    parser.add_argument('--simulation', action='store_true',
                        help='Use simulation mode for XADD compilation')
    args = parser.parse_args()
    test_xadd(env_name=args.env,
              cpf=args.cpf,
              save_graph=args.save_graph,
              simulation=args.simulation)



from pyRDDLGym.Visualizer.StateViz import StateViz
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
import matplotlib


class ChartVisualizer(StateViz):

    def __init__(self, model: PlanningModel,
                 steps_history=None,
                 figure_size=[10, 10],
                 dpi=100,
                 fontsize=10,
                 markersize=4,
                 boolcol=['red', 'green'],
                 loccol='black',
                 scale_full_history: bool = True,
                 ranges=None) -> None:
        self._model = model
        self._figure_size = figure_size
        self._dpi = dpi
        self._fontsize = fontsize
        self._markersize = markersize
        self._boolcol = boolcol
        self._loccol = loccol
        self._scale_full_history = scale_full_history
        self._ranges = ranges

        self._fig, self._ax = None, None
        self._data = None
        self._img = None

        if steps_history is None:
            steps_history = model.horizon
        self._steps = steps_history
        self._state_hist = {}
        self._state_shapes = {}
        self._labels = {}
        self._historic_min_max = {}
        for (state, values) in self._model.states.items():
            values = np.atleast_1d(values)
            self._state_hist[state] = np.full(shape=(len(values), self._steps),
                                              fill_value=np.nan)
            self._state_shapes[state] = self._model.object_counts(
                self._model.param_types[state])
            self._labels[state] = list(map(
                ','.join, 
                self._model.variations(self._model.param_types[state])
            ))
            self._historic_min_max[state] = (np.inf, -np.inf)
        self._step = 0
        self._drawn_artists = False
    
    def _setup(self):
        
        # prepare plot
        self._fig, self._ax = plt.subplots(len(self._state_hist), 1,
                                           squeeze=True,
                                           figsize=self._figure_size,
                                           sharex=True)
        if len(self._state_hist) == 1:
            self._ax = (self._ax,)
        self._fig.canvas.draw()
        
        # prepare artists
        self._lines = {}
        for (y, (state, values)) in enumerate(self._state_hist.items()):
            values = values[::-1, :]
            self._lines[state] = []
            
            # title, labels and cursor line
            self._ax[y].xaxis.label.set_fontsize(self._fontsize)
            self._ax[y].yaxis.label.set_fontsize(self._fontsize)
            if y == len(self._state_hist) - 1:
                self._ax[y].set_xlabel('decision epoch')
            self._ax[y].set_ylabel(state, rotation=15)
            
            # plot boolean fluent
            if self._model.variable_ranges[state] == 'bool':
                line = self._ax[y].pcolormesh(values,
                                              edgecolors=self._loccol,
                                              linewidth=0.5,
                                              cmap=matplotlib.colors.ListedColormap(self._boolcol),
                                              vmin=0, 
                                              vmax=1)
                self._lines[state].append(line)
                
                # set the legend colors
                patches = [
                    matplotlib.patches.Patch(color=self._boolcol[0], label='false'),
                    matplotlib.patches.Patch(color=self._boolcol[1], label='true')
                ]
                self._ax[y].legend(handles=patches, loc='upper right')
                
                # set the state variables to the y-axis labels
                labels = self._labels[state]
                self._ax[y].yaxis.set_ticks([0, len(labels)])
                self._ax[y].yaxis.set(ticks=np.arange(0.5, len(labels), 1),
                                      ticklabels=labels)
                self._ax[y].set_yticklabels(labels,
                                            fontdict={"fontsize": self._fontsize},
                                            rotation=30)
            
            # plot real fluent
            elif self._model.variable_ranges[state] == 'real':
                for (i, var) in enumerate(self._labels[state]):
                    line = self._ax[y].plot(values[i, :],
                                            'o-',
                                            markersize=self._markersize,
                                            label=var)[0]
                    self._lines[state].append(line)
                self._ax[y].set_xlim([0, values[i, :].size])
                self._ax[y].legend(loc='upper right')
            
            # plot int and object fluent
            else:
                for (i, var) in enumerate(self._labels[state]):
                    line = self._ax[y].plot(np.arange(values[i, :].size),
                                            values[i, :],
                                            linestyle=':',
                                            marker='o',
                                            markersize=self._markersize,
                                            markerfacecolor='none',
                                            label=var)[0]
                    self._lines[state].append(line)
                self._ax[y].set_xlim([0, values[i, :].size])
                if var != '':
                    self._ax[y].legend(loc='upper right')
        plt.tight_layout()
                    
        self._backgrounds = [self._fig.canvas.copy_from_bbox(ax.bbox) for ax in self._ax]
        self._drawn_artists = True
        
    def convert2img(self, fig, ax):
        fig.canvas.draw()

        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        img = Image.fromarray(data)

        self._data = data
        self._img = img
        return img

    def render(self, state):
        if not self._drawn_artists:
            self._setup()

        # update the state info
        if self._step >= self._steps:
            for values in self._state_hist.values():
                values[:, :-1] = values[:, 1:]
        states = {name: np.full(shape=shape, fill_value=np.nan)
                  for (name, shape) in self._state_shapes.items()}
        for (name, value) in state.items():
            var, objects = self._model.parse(name)
            states[var][self._model.indices(objects)] = value
        index = min(self._step, self._steps - 1)
        for (name, values) in states.items():
            self._state_hist[name][:, index] = np.ravel(values, order='C')

        # draw color plots
        for (y, (state, values)) in enumerate(self._state_hist.items()):
            values = values[::-1, :]

            # scaling of y axis
            vmin, vmax = None, None
            if self._scale_full_history:
                hmin, hmax = self._historic_min_max[state]
                valid_values = values[np.isfinite(values)]
                vmin = min(hmin, np.min(valid_values))
                vmax = max(hmax, np.max(valid_values))
                self._historic_min_max[state] = (vmin, vmax)
            if self._ranges is not None and state in self._ranges:
                vmin, vmax = self._ranges[state]

            # plot boolean fluent
            if self._model.variable_ranges[state] == 'bool':
                self._fig.canvas.restore_region(self._backgrounds[y])
                self._lines[state][0].set_array(values)
                self._ax[y].draw_artist(self._lines[state][0])

            # plot real fluent
            elif self._model.variable_ranges[state] == 'real':
                self._fig.canvas.restore_region(self._backgrounds[y])
                for (i, var) in enumerate(self._labels[state]):
                    self._lines[state][i].set_ydata(values[i, :])
                    self._ax[y].draw_artist(self._lines[state][i])
                self._ax[y].set_ylim([vmin, vmax])
                
            # plot int and object fluent
            else:
                
                # offset each curve by a small amount in x so all can be seen
                num_ser = len(self._labels[state])
                if num_ser > 1:
                    offsets = np.linspace(-0.05, 0.05, num=num_ser, endpoint=True)
                else:
                    offsets = np.zeros((num_ser,))
                
                self._fig.canvas.restore_region(self._backgrounds[y])
                for (i, var) in enumerate(self._labels[state]):
                    self._lines[state][i].set_xdata(np.arange(values[i, :].size) + offsets[i])
                    self._lines[state][i].set_ydata(values[i, :])
                    self._ax[y].draw_artist(self._lines[state][i])
                self._ax[y].set_ylim([vmin, vmax])
            self._fig.canvas.blit(self._ax[y].bbox)

        self._step = self._step + 1

        img = self.convert2img(self._fig, self._ax)

        return img



import os

from pyRDDLGym import RDDLEnv
from pyRDDLGym import ExampleManager
from pyRDDLGym.Core.Policies.Agents import RandomAgent

ENVS = ['HVAC', 'RaceCar', 'UAV_continuous', 'MarsRover', 'PowerGen_continuous', 'MountainCar', 'RecSim', 'Reservoir_continuous']

###################
# Competition run instances
###################
INST_NAMES = ['0c', '3c', '5c']
HVAC_PARAMS = [
    # difficulty is controlled by the number of zones + heaters (e.g. scale)
    # and the switching of occupancy affects more zones (this should make it
    # difficult for determinization methods)
    {'num_heaters': 3, 'num_zones': 2, 'density': 0.2,
     'temp-zone-range-init': (0., 15.), 'temp-heater-range-init': (0., 10.),
     'TEMP-ZONE-MIN': 22.0, 'TEMP-ZONE-MAX': 25.0,
     'p-switch-number': 0, 'p-switch-prob': 0.02,
     'horizon': 100, 'discount': 1.0},
    {'num_heaters': 5, 'num_zones': 5, 'density': 0.2,
     'temp-zone-range-init': (0., 15.), 'temp-heater-range-init': (0., 10.),
     'TEMP-ZONE-MIN': 22.0, 'TEMP-ZONE-MAX': 25.0,
     'p-switch-number': 2, 'p-switch-prob': 0.02,
     'horizon': 100, 'discount': 1.0},
    {'num_heaters': 15, 'num_zones': 20, 'density': 0.2,
     'temp-zone-range-init': (0., 15.), 'temp-heater-range-init': (0., 10.),
     'TEMP-ZONE-MIN': 22.0, 'TEMP-ZONE-MAX': 25.0,
     'p-switch-number': 15, 'p-switch-prob': 0.02,
     'horizon': 100, 'discount': 1.0}

]
MarsRover_PARAMS = [
    # the difficulty here is modulated by increasing number of minerals and rovers.
    # some problems have too many rovers, and some less rovers than minerals.
    # the agent is forced to delay some small immediate reward for taking
    # some additional time steps and fuel to accept a much larger reward in future.
    {'num_minerals': 1, 'num_rovers': 2, 'location_bounds': (-10., 10.),
     'area_bounds': (1., 5.), 'value_bounds': (0., 20.),
     'horizon': 100, 'discount': 1.0},
    {'num_minerals': 7, 'num_rovers': 6, 'location_bounds': (-10., 10.),
     'area_bounds': (1., 3.), 'value_bounds': (0., 20.),
     'horizon': 100, 'discount': 1.0},
    {'num_minerals': 20, 'num_rovers': 5, 'location_bounds': (-10., 10.),
     'area_bounds': (1., 2.), 'value_bounds': (0., 20.),
     'horizon': 100, 'discount': 1.0}
]
MountainCar_PARAMS = [
    {'terrain_xleft': -1.2, 'terrain_widths': [0.7, 1.1], 'terrain_heights': [0.9, 0.9],
     'num_points': 100, 'pos': -0.6, 'vel': 0.01, 'goal-min': 0.5, 'horizon': 200, 'discount': 1.0},
    {'terrain_xleft': -1.2, 'terrain_widths': [0.7, 2.3, 2.1, 1.1], 'terrain_heights': [0.9, 0.8, 0.7, 1.0],
     'num_points': 100, 'pos': -0.6, 'vel': 0.01, 'goal-min': 4.9, 'horizon': 200, 'discount': 1.0},
    {'terrain_xleft': -1.2, 'terrain_widths': [0.7, 2.3, 2.2, 2.1, 1.1], 'terrain_heights': [0.5, 0.8, 0.9, 1.0, 1.0],
     'num_points': 100, 'pos': -0.6, 'vel': 0.01, 'goal-min': 7.1, 'horizon': 200, 'discount': 1.0}
]
PowerGen_continuous_PARAMS = [
    # difficulty is controlled by number and diversity of plant types, many of
    # which progressively require unit commitment
    {'num_gas': 2, 'num_nuclear': 0, 'num_solar': 0,
     'demand_scale': 1.0, 'temp_variance': 5.0, 'temp_range': (-30.0, 40.0),
     'horizon': 100, 'discount': 1.0},
    {'num_gas': 2, 'num_nuclear': 2, 'num_solar': 1,
     'demand_scale': 3.0, 'temp_variance': 7.0, 'temp_range': (-30.0, 40.0),
     'horizon': 100, 'discount': 1.0},
    {'num_gas': 3, 'num_nuclear': 4, 'num_solar': 4,
     'demand_scale': 6.0, 'temp_variance': 9.0, 'temp_range': (-30.0, 40.0),
     'horizon': 100, 'discount': 1.0}
]
RaceCar_PARAMS = [
    # difficulty is controlled by the number of obstacles, as well as the radius
    # of the goal region, which is shrinking
    {'num_blocks': 1, 'min_block_size': 0.05, 'max_block_size': 0.5, 'scale': 1.0,
     'goal_radius': 0.06, 'horizon': 150, 'discount': 1.0},
    {'num_blocks': 3, 'min_block_size': 0.05, 'max_block_size': 0.4, 'scale': 1.5,
     'goal_radius': 0.04, 'horizon': 150, 'discount': 1.0},
    {'num_blocks': 5, 'min_block_size': 0.05, 'max_block_size': 0.3, 'scale': 2.0,
     'goal_radius': 0.02, 'horizon': 150, 'discount': 1.0}
]
RecSim_PARAMS = [
    # Difficulty is controlled by the number users and number of providers (providers = num_provider_clusters*provider_fan_out) and their dispersity
    # and providers.
    {'provider_dispersion': 10.0, 'provider_fan_out': 2, 'num_provider_clusters': 3, 'num_users': 10,
     'docs_per_cluster': 5, 'user_stddev': 20.0, 'horizon': 100, 'discount': 1.0 },
    {'provider_dispersion': 62.0, 'provider_fan_out': 5, 'num_provider_clusters': 20, 'num_users': 90,
     'docs_per_cluster': 5, 'user_stddev': 20.0, 'horizon': 100, 'discount': 1.0 },
    {'provider_dispersion': 62.0, 'provider_fan_out': 10, 'num_provider_clusters': 100, 'num_users': 800,
     'docs_per_cluster': 5,  'user_stddev': 20.0, 'horizon': 100, 'discount': 1.0 }
]
Reservoir_continuous_PARAMS = [
    # the number of reservoirs, amount of interdependence between them,
    # the rain variance and the narrowing of the target water level ranges all
    # contribute to difficulty
    {'num_reservoirs': 2, 'max_edges': 1,
     'top_range': (100., 200.), 'target_range': (0.4, 0.7), 'rain_var': 5.,
     'horizon': 100, 'discount': 1.0},
    {'num_reservoirs': 10, 'max_edges': 5,
     'top_range': (100., 400.), 'target_range': (0.2, 0.5), 'rain_var': 20.,
     'horizon': 100, 'discount': 1.0},
    {'num_reservoirs': 30, 'max_edges': 10,
     'top_range': (100., 600.), 'target_range': (0.1, 0.2), 'rain_var': 50.,
     'horizon': 100, 'discount': 1.0}
]
UAV_continuous_PARAMS = [
    # the difficulty of this problem is the noise of uncontrollable aircraft,
    # so the difficulty is controlled by number of such craft (relative to controlled),
    # the total number of craft (scale) and the variance of the uncontrolled state
    {'num_aircraft': 1, 'num_control': 1, 'variance': 1.0,
     'xrange': (-50., 50.), 'yrange': (-50., 50.), 'zrange': (0., 100.),
     'horizon': 100, 'discount': 1.0},
    {'num_aircraft': 3, 'num_control': 2, 'variance': 4.0,
     'xrange': (-50., 50.), 'yrange': (-50., 50.), 'zrange': (0., 100.),
     'horizon': 100, 'discount': 1.0},
    {'num_aircraft': 40, 'num_control': 10, 'variance': 10.0,
     'xrange': (-50., 50.), 'yrange': (-50., 50.), 'zrange': (0., 100.),
     'horizon': 100, 'discount': 1.0}
]

INST_PARAMS = {
    'HVAC' : HVAC_PARAMS,
    'RaceCar' : RaceCar_PARAMS,
    'UAV_continuous' : UAV_continuous_PARAMS,
    'MarsRover' : MarsRover_PARAMS,
    'PowerGen_continuous' : PowerGen_continuous_PARAMS,
    'MountainCar' : MountainCar_PARAMS,
    'RecSim' : RecSim_PARAMS,
    'Reservoir_continuous' : Reservoir_continuous_PARAMS
}


def Test_instance(EnvInfo, inst_path):
    myEnv = RDDLEnv.RDDLEnv(domain=EnvInfo.get_domain(),
                            instance=inst_path,
                            enforce_action_constraints=False,
                            debug=False)
    myEnv.set_visualizer(EnvInfo.get_visualizer())
    agent = RandomAgent(action_space=myEnv.action_space,
                        num_actions=myEnv.numConcurrentActions)

    total_reward = 0
    state = myEnv.reset()
    for step in range(myEnv.horizon):
        myEnv.render()
        action = agent.sample_action()
        next_state, reward, done, info = myEnv.step(action)
        total_reward += reward
        print()
        print(f'step       = {step}')
        print(f'state      = {state}')
        print(f'action     = {action}')
        print(f'next state = {next_state}')
        print(f'reward     = {reward}')
        state = next_state
        if done:
            break
    print(f'episode ended with reward {total_reward}')
    myEnv.close()


def main(env, params, name):
    print(env, params, name)
    inst_path = None

    path = os.path.dirname(os.path.abspath(__file__))
    path = os.path.join(path, 'Instances', env)


    EnvInfo = ExampleManager.GetEnvInfo(env)
    inst_path = EnvInfo.generate_instance(name, params, path)

    if not inst_path:
        inst_path = os.path.join(path, 'instance'+name+'.rddl')
    Test_instance(EnvInfo, inst_path)






if __name__ == "__main__":
    inst_index = 0
    env = ENVS[5]
    inst_name = INST_NAMES[inst_index]
    inst_params = INST_PARAMS[env][inst_index]

    main(env, inst_params, inst_name)



from pyRDDLGym.Core.Env import RDDLEnv as RDDLEnv
from pyRDDLGym.Core.Policies.Agents import RandomAgent
import math

FOLDER = 'pyRDDLGym/Examples/MarsRover/'


def mars_rover_update(state, action):
    power_d1 = math.sqrt(action['power-x_d1'] ** 2 + action['power-y_d1'] ** 2)
    power_d2 = math.sqrt(action['power-x_d2'] ** 2 + action['power-y_d2'] ** 2)
    if power_d1 >= 0.05:
        u_x_d1 = 0.05 * action['power-x_d1'] / power_d1
    else:
        u_x_d1 = action['power-x_d1']
    if power_d2 >= 0.05:
        u_x_d2 = 0.05 * action['power-x_d2'] / power_d2
    else:
        u_x_d2 = action['power-x_d2']
    if power_d1 >= 0.05:
        u_y_d1 = 0.05 * action['power-y_d1'] / power_d1
    else:
        u_y_d1 = action['power-y_d1']
    if power_d2 >= 0.05:
        u_y_d2 = 0.05 * action['power-y_d2'] / power_d2
    else:
        u_y_d2 = action['power-y_d2']

    next_interm = {'power_d1': power_d1,
                   'power_d2': power_d2,
                   'u-x_d1': u_x_d1,
                   'u-y_d1': u_y_d1,
                   'u-x_d2': u_x_d2,
                   'u-y_d2': u_y_d2}

    new_vel_x_d1 = state['vel-x_d1'] + 0.1 * u_x_d1
    new_vel_y_d1 = state['vel-y_d1'] + 0.1 * u_y_d1
    new_vel_x_d2 = state['vel-x_d2'] + 0.1 * u_x_d2
    new_vel_y_d2 = state['vel-y_d2'] + 0.1 * u_y_d2
    new_pos_x_d1 = state['pos-x_d1'] + 0.1 * state['vel-x_d1']
    new_pos_y_d1 = state['pos-y_d1'] + 0.1 * state['vel-y_d1']
    new_pos_x_d2 = state['pos-x_d2'] + 0.1 * state['vel-x_d2']
    new_pos_y_d2 = state['pos-y_d2'] + 0.1 * state['vel-y_d2']

    print(action)
    new_mineral_harvested_m1 = state['mineral-harvested_m1'] or (
        (math.sqrt((state['pos-x_d1'] - 5) ** 2 +
         (state['pos-y_d1'] - 5) ** 2) < 6 and action['harvest_d1'])
        or (math.sqrt((state['pos-x_d2'] - 5) ** 2 + (state['pos-y_d2'] - 5) ** 2) < 6 and action['harvest_d2'])
    )
    new_mineral_harvested_m2 = state['mineral-harvested_m2'] or (
        (math.sqrt((state['pos-x_d1'] - (-8)) ** 2 +
         (state['pos-y_d1'] - (-8)) ** 2) < 8 and action['harvest_d1'])
        or (math.sqrt((state['pos-x_d2'] - (-8)) ** 2 + (state['pos-y_d2'] - (-8)) ** 2) < 8 and action['harvest_d2'])
    )

    next_state = {'vel-x_d1': new_vel_x_d1,
                  'vel-y_d1': new_vel_y_d1,
                  'pos-x_d1': new_pos_x_d1,
                  'pos-y_d1': new_pos_y_d1,
                  'vel-x_d2': new_vel_x_d2,
                  'vel-y_d2': new_vel_y_d2,
                  'pos-x_d2': new_pos_x_d2,
                  'pos-y_d2': new_pos_y_d2,
                  'mineral-harvested_m1': new_mineral_harvested_m1,
                  'mineral-harvested_m2': new_mineral_harvested_m2}

    mineral_m1_cond = (
        8 if (
            (math.sqrt((state['pos-x_d1'] - 5) ** 2 +
             (state['pos-y_d1'] - 5) ** 2) < 6 and action['harvest_d1'])
            or (math.sqrt((state['pos-x_d2'] - 5) ** 2 + (state['pos-y_d2'] - 5) ** 2) < 6 and action['harvest_d2'])
        )
        else 0
    )

    mineral_m2_cond = (
        5 if (
            (math.sqrt((state['pos-x_d1'] - -8) ** 2 +
             (state['pos-y_d1'] - -8) ** 2) < 8 and action['harvest_d1'])
            or (math.sqrt((state['pos-x_d2'] - -8) ** 2 + (state['pos-y_d2'] - -8) ** 2) < 8 and action['harvest_d2'])
        )
        else 0
    )

    reward = -(u_x_d1 ** 2 + u_y_d1 ** 2) - (u_x_d2 ** 2 + u_y_d2 ** 2) + \
        mineral_m1_cond + mineral_m2_cond \
        - 1. * action['harvest_d1'] - 1. * action['harvest_d2']
    return reward, next_state, next_interm


def main():
    myEnv = RDDLEnv.RDDLEnv(domain=FOLDER + 'domain.rddl',
                            instance=FOLDER + 'instance0.rddl')
    agent = RandomAgent(action_space=myEnv.action_space,
                        num_actions=myEnv.numConcurrentActions)

    from pprint import pprint
    pprint(vars(myEnv.model))
    # print(myEnv.model.reward)

    print(myEnv.model.cpfs['mineral-harvested_m1\''])
    print(myEnv.model.cpfs['mineral-harvested_m2\''])
    total_reward = 0
    state = myEnv.reset()
    test_state = state
    print(state)
    for step in range(myEnv.model.horizon):
        myEnv.render()
        action = agent.sample_action()
        action['harvest_d1'] = bool(action['harvest_d1'])
        action['harvest_d2'] = bool(action['harvest_d2'])
        next_state, reward, done, info = myEnv.step(action)
        total_reward += reward
        print("\nstep = {}".format(step))
        print('reward = {}'.format(reward))
        print('state = {}'.format(state))
        print('action = {}'.format(action))
        print('next_state = {}'.format(next_state))
        print('interm = {}'.format(myEnv.model.interm))
        state = next_state

        test_reward, next_test_state, test_interm = mars_rover_update(
            test_state, action)
        assert test_reward == reward
        assert next_test_state == next_state
        assert test_interm == myEnv.model.interm, str(
            test_interm) + '\n' + str(myEnv.model.interm)
        test_state = next_test_state

    print("episode ended with reward {}".format(total_reward))


if __name__ == "__main__":
    main()



import configparser
import csv
import os
import re

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLEnvironmentNotExist, RDDLInstanceNotExist
# from pyRDDLGym.Examples.InstanceGenerators.UAVInstanceGenerator import UAVInstanceGenerator

HEADER = ['name', 'description', 'location', 'instances', 'viz', 'generator']


def rebuild():
    path = os.path.dirname(os.path.abspath(__file__))    
    # path_to_manifest = os.path.join(path, '../Domains/manifest.csv')
    path_to_manifest = os.path.join(path, 'manifest.csv')
    print (path, path_to_manifest)
    
    with open(path_to_manifest, 'w', newline='') as file:
        
        # write the header for the manifest
        writer = csv.writer(file, delimiter=',')
        writer.writerow(HEADER)
        
        # walk through current folder to find valid domains
        for dirpath, _, filenames in os.walk(path):
            if 'domain.info' in filenames:
                infopath = os.path.join(dirpath, 'domain.info')
                config = configparser.RawConfigParser()
                config.optionxform = str 
                config.read(infopath)
                general = dict(config.items('General'))
                name = general.get('name', None)
                desc = general.get('description', None)
                viz = general.get('viz', None)
                generator=general.get('generator', None)
                loc = dirpath[len(path):]
                loc = loc.replace('\\', '/') + '/'
                instances = [fname[8:-5] for fname in filenames 
                             if fname.startswith('instance') and fname.endswith('.rddl')]
                instances = ','.join(instances)
                writer.writerow([name, desc, loc, instances, viz, generator])


def load():
    path = os.path.dirname(os.path.abspath(__file__))
    # path_to_manifest = os.path.join(path, '../Domains/manifest.csv')
    path_to_manifest = os.path.join(path, 'manifest.csv')
    if not os.path.isfile(path_to_manifest):
        return {}
        
    EXP_DICT = {}
    with open(path_to_manifest) as file:
        reader = csv.reader(file, delimiter=',')
        for i, row in enumerate(reader):
            if i > 0:
                name, *entries = row
                EXP_DICT[name] = dict(zip(HEADER[1:], entries))
    return EXP_DICT


if __name__ == '__main__':
    rebuild()


class ExampleManager:
    
    EXP_DICT = load()
    
    def __init__(self, env: str):
        self.env = env
        
        if not ExampleManager.EXP_DICT:
            ExampleManager.RebuildExamples()
            
        if env not in ExampleManager.EXP_DICT:
            raise RDDLEnvironmentNotExist(
                f'Environment <{env}> does not exist, '
                f'must be one of {set(ExampleManager.EXP_DICT.keys())}.')

        self.path_to_env = os.path.dirname(os.path.abspath(__file__)) + \
                            ExampleManager.EXP_DICT[env]['location']

    def get_domain(self):
        return self.path_to_env + 'domain.rddl'

    def list_instances(self):
        files = os.listdir(self.path_to_env)
        instances = []
        for file in files:
            x = re.search("instance\d+.*", file)
            if x is not None:
                instances.append(file)
        return instances

    def get_instance(self, num: str):
        instance = f'instance{num}.rddl'
        if not os.path.exists(self.path_to_env + instance):
            raise RDDLInstanceNotExist(
                f'instance {instance} does not exist for '
                f'example environment <{self.env}>.')
        return self.path_to_env + instance

    def get_visualizer(self):
        viz = None
        viz_info = ExampleManager.EXP_DICT[self.env]['viz']
        if viz_info:
            module, viz_class_name = viz_info.strip().split('.')
            viz_package_name = 'pyRDDLGym.Visualizer.' + module
            viz_package = __import__(viz_package_name, {}, {}, viz_class_name)
            viz = getattr(viz_package, viz_class_name)
        return viz

    def generate_instance(self, name, params, path=None):
        if path is None:
            path = self.path_to_env

        generator_info = ExampleManager.EXP_DICT[self.env]['generator']
        if generator_info == '':
            return None

        module, generator_class_name = generator_info.strip().split('.')
        generator_package_name = 'pyRDDLGym.Examples.InstanceGenerators.' + module
        generator_package = __import__(generator_package_name, {}, {}, generator_class_name)
        generator = getattr(generator_package, generator_class_name)

        generator = generator()
        generator.save_instance(name, params, path)
        return os.path.join(path, f'instance{name}.rddl')





    @staticmethod
    def ListExamples():
        print('Available example environment(s):')
        for key, values in ExampleManager.EXP_DICT.items():
            print(key + ' -> ' + values['description'])
    
    @staticmethod
    def RebuildExamples():
        rebuild()
        ExampleManager.EXP_DICT = load()
        ExampleManager.ListExamples()
        
    @staticmethod
    def GetEnvInfo(env):
        return ExampleManager(env)
    



import copy
import numpy as np

from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidObjectError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLMissingCPFDefinitionError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLRepeatedVariableError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLValueOutOfRangeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedCPFError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel


class RDDLLiftedModel(PlanningModel):
    '''Represents a RDDL domain + instance in lifted form.'''
    
    PRIMITIVE_TYPES = {
        'int': int,
        'real': float,
        'bool': bool
    }
    
    def __init__(self, rddl):
        super(RDDLLiftedModel, self).__init__()
        
        self.SetAST(rddl)
        
        self._extract_objects()            
        self._extract_variable_information()  
                
        self._extract_states()
        self._extract_actions()
        self._extract_derived_and_interm()
        self._extract_observ()
        self._extract_non_fluents()
        
        self.reward = self._AST.domain.reward
        self._extract_cpfs()
        self._extract_constraints()
        
        self._extract_horizon()
        self._extract_discount()
        self._extract_max_actions()
        
    def _extract_objects(self):
        
        # objects of each type as defined in the non-fluents {..} block
        ast_objects = self._AST.non_fluents.objects
        if not ast_objects or ast_objects[0] is None:
            ast_objects = []
        ast_objects = dict(ast_objects)
        
        # record the set of objects of each type defined in the domain
        objects, objects_rev, enums = {}, {}, set() 
        for (name, pvalues) in self._AST.domain.types:
            
            # check duplicated type
            if name in objects:
                raise RDDLInvalidObjectError(
                    f'Type <{name}> is repeated in types block.')
            
            # instance object
            if pvalues == 'object': 
                objects[name] = ast_objects.get(name, None)
                if objects[name] is None:
                    raise RDDLInvalidObjectError(
                        f'Type <{name}> has no objects defined in the instance.')
                objects[name] = list(map(self.object_name, objects[name]))
                
            # domain object
            else: 
                objects[name] = list(map(self.object_name, pvalues))
                enums.add(name)
        
            # make sure types do not share an object - record type of each object
            for obj in objects[name]:
                if obj in objects_rev:
                    if objects_rev[obj] == name:
                        raise RDDLInvalidObjectError(
                            f'Type <{name}> contains duplicated object <{obj}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'Types <{name}> and <{objects_rev[obj]}> '
                            f'can not share the same object <{obj}>.')
                objects_rev[obj] = name
        
        # check that all types in instance are declared in domain
        for ptype in ast_objects:
            if ptype not in objects:
                raise RDDLInvalidObjectError(
                    f'Type <{ptype}> defined in the instance is not declared in '
                    f'the domain.')
        
        # maps each object to its canonical order as it appears in definition
        self.index_of_object = {obj: i 
                                for objs in objects.values() 
                                    for (i, obj) in enumerate(objs)}
        
        self.objects, self.objects_rev, self.enums = objects, objects_rev, enums
    
    def _extract_variable_information(self):
        
        # extract some basic information about variables in the domain:
        # 1. object parameters needed to evaluate
        # 2. type (e.g. state-fluent, action-fluent)
        # 3. range (e.g. int, real, bool, type)
        # 4. save default values
        var_params, var_types, var_ranges, default_values = {}, {}, {}, {}
        for pvar in self._AST.domain.pvariables: 
            
            # make sure variable is not defined more than once        
            primed_name = name = pvar.name
            if name in var_params:
                raise RDDLRepeatedVariableError(
                    f'{pvar.fluent_type} <{name}> has the same name as '
                    f'another {var_types[name]} variable.')
                
            # make sure name does not contain separators
            SEPARATORS = [PlanningModel.FLUENT_SEP, PlanningModel.OBJECT_SEP] 
            for separator in SEPARATORS:
                if separator in name:
                    raise RDDLInvalidObjectError(
                        f'Variable name <{name}> contains an '
                        f'illegal separator {separator}.')
            
            # record its type, parameters and range
            if pvar.is_state_fluent():
                primed_name = name + PlanningModel.NEXT_STATE_SYM
            ptypes = pvar.param_types
            if ptypes is None:
                ptypes = []
            var_params[name] = var_params[primed_name] = ptypes        
            var_types[name] = pvar.fluent_type
            if pvar.is_state_fluent():
                var_types[primed_name] = 'next-state-fluent'                
            var_ranges[name] = var_ranges[primed_name] = pvar.range    
            
            # save default value
            default_values[name] = self._extract_default_value(pvar)
            
        # maps each variable (as appears in RDDL) to list of grounded variations
        self.grounded_names = {var: list(self.ground_names(var, types))
                               for (var, types) in var_params.items()}        
        
        self.param_types, self.variable_types, self.variable_ranges = \
            var_params, var_types, var_ranges
        self.default_values = default_values
    
    def _grounded_dict_to_dict_of_list(self, grounded_dict):
        new_dict = {}
        for (var, values_dict) in grounded_dict.items():
            grounded_names = list(values_dict.values())
            if self.param_types[var]:
                new_dict[var] = grounded_names
            else:
                assert len(grounded_names) == 1
                new_dict[var] = grounded_names[0]
        return new_dict
    
    def _extract_default_value(self, pvar):
        prange, default = pvar.range, pvar.default
        if default is not None:
            
            # is an object
            if isinstance(default, str):
                default = self.object_name(default)
                if default not in self.objects.get(prange, set()):
                    raise RDDLTypeError(
                        f'Default value {default} of variable <{pvar.name}> '
                        f'is not an object of type <{prange}>.')                     
            
            # is a primitive
            else:
                dtype = RDDLLiftedModel.PRIMITIVE_TYPES.get(prange, None)
                if dtype is None:
                    raise RDDLTypeError(
                        f'Type <{prange}> of variable <{pvar.name}> is an object '
                        f'or enumerated type, but assigned a default value '
                        f'{default}.')
                
                # type cast
                if not np.can_cast(default, dtype):
                    raise RDDLTypeError(
                        f'Default value {default} of variable <{pvar.name}> '
                        f'cannot be cast to required type <{prange}>.')
                default = dtype(default)
            
        return default
    
    def _extract_states(self):
        
        # get the information for each state from the domain
        PRIME = PlanningModel.NEXT_STATE_SYM
        states, statesranges, nextstates, prevstates = {}, {}, {}, {}
        for pvar in self._AST.domain.pvariables:
            if pvar.is_state_fluent():
                name, ptypes = pvar.name, pvar.param_types
                statesranges[name] = pvar.range
                nextstates[name] = name + PRIME
                prevstates[name + PRIME] = name
                default = self.default_values[name]           
                states[name] = {gname: default 
                                for gname in self.ground_names(name, ptypes)} 
                
        # update the state values with the values in the instance
        initstates = copy.deepcopy(states)
        init_state_info = getattr(self._AST.instance, 'init_state', [])
        for ((name, params), value) in init_state_info:
                
            # check whether name is a valid state-fluent
            if name not in initstates:
                raise RDDLUndefinedVariableError(
                    f'Variable <{name}> referenced in init-state block '
                    f'is not a valid state-fluent.')
                    
            # extract the grounded name and check that parameters are valid
            if params is not None:
                params = list(map(self.object_name, params))
            gname = self.ground_name(name, params)
            if gname not in initstates[name]:
                raise RDDLInvalidObjectError(
                    f'Parameter(s) {params} of state-fluent <{name}> '
                    f'declared in the init-state block are not valid.')
                
            # make sure value is correct type
            if isinstance(value, str):
                value = self.object_name(value)
                value_type = self.objects_rev.get(value, None)
                required_type = statesranges[name]
                if value_type != required_type:
                    if value_type is None:
                        raise RDDLInvalidObjectError(
                            f'State-fluent <{name}> of type <{required_type}> '
                            f'is initialized in init-state block with undefined '
                            f'object <{value}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'State-fluent <{name}> of type <{required_type}> '
                            f'is initialized in init-state block with object '
                            f'<{value}> of type {value_type}.')
                        
            initstates[name][gname] = value
                
        # state dictionary associates the variable lifted name with a list of
        # values for all variations of parameter arguments in C-based order
        # if the fluent does not have parameters, then the value is a scalar
        states = self._grounded_dict_to_dict_of_list(states)
        initstates = self._grounded_dict_to_dict_of_list(initstates)     
        
        self.states, self.statesranges = states, statesranges
        self.next_state, self.prev_state = nextstates, prevstates
        self.init_state = initstates
    
    def _value_list_or_scalar_from_default(self, pvar):
        default = self.default_values[pvar.name]
        ptypes = pvar.param_types   
        if ptypes is None:
            return default
        else:
            num_variations = 1
            for ptype in ptypes:
                num_variations *= len(self.objects[ptype])
            return [default] * num_variations
    
    def _extract_actions(self):
        
        # actions are stored similar to states described above
        actions, actionsranges = {}, {}
        for pvar in self._AST.domain.pvariables:
            if pvar.is_action_fluent():
                actionsranges[pvar.name] = pvar.range
                actions[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.actions, self.actionsranges = actions, actionsranges
    
    def _extract_derived_and_interm(self):
        
        # derived and interm are stored similar to states described above
        derived, interm = {}, {}
        for pvar in self._AST.domain.pvariables:
            if pvar.is_derived_fluent():
                derived[pvar.name] = self._value_list_or_scalar_from_default(pvar)
            elif pvar.is_intermediate_fluent():
                interm[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.derived, self.interm = derived, interm
    
    def _extract_observ(self):
        
        # observed are stored similar to states described above
        observ, observranges = {}, {}
        for pvar in self._AST.domain.pvariables:
            if pvar.is_observ_fluent():
                observranges[pvar.name] = pvar.range
                observ[pvar.name] = self._value_list_or_scalar_from_default(pvar)
        self.observ, self.observranges = observ, observranges
        
    def _extract_non_fluents(self):
        
        # extract non-fluents values from the domain defaults
        non_fluents = {}
        for pvar in self._AST.domain.pvariables:
            if pvar.is_non_fluent():
                name, ptypes = pvar.name, pvar.param_types
                default = self.default_values[name] 
                non_fluents[name] = {gname: default
                                     for gname in self.ground_names(name, ptypes)}
        
        # update non-fluent values with the values in the instance
        non_fluent_info = getattr(self._AST.non_fluents, 'init_non_fluent', [])
        for ((name, params), value) in non_fluent_info:
                
            # check whether name is a valid non-fluent
            grounded_names = non_fluents.get(name, None)
            if grounded_names is None:
                raise RDDLUndefinedVariableError(
                    f'Variable <{name}> referenced in non-fluents block '
                    f'is not a valid non-fluent.')
                
            # extract the grounded name and check that parameters are valid
            if params is not None:
                params = list(map(self.object_name, params))
            gname = self.ground_name(name, params)                           
            if gname not in grounded_names:
                raise RDDLInvalidObjectError(
                    f'Parameter(s) {params} of non-fluent <{name}> '
                    f'as declared in the non-fluents block are not valid.')
                    
            # make sure value is correct type
            if isinstance(value, str):
                value = self.object_name(value)
                value_type = self.objects_rev.get(value, None)
                required_type = self.variable_ranges[name]
                if value_type != required_type:
                    if value_type is None:
                        raise RDDLInvalidObjectError(
                            f'Non-fluent <{name}> of type <{required_type}> '
                            f'is initialized in non-fluents block with '
                            f'undefined object <{value}>.')
                    else:
                        raise RDDLInvalidObjectError(
                            f'Non-fluent <{name}> of type <{required_type}> '
                            f'is initialized in non-fluents block with object '
                            f'<{value}> of type <{value_type}>.')
                        
            grounded_names[gname] = value
                                        
        # non-fluents are stored similar to states described above
        self.nonfluents = self._grounded_dict_to_dict_of_list(non_fluents)
    
    def _extract_cpfs(self):
        cpfs = {}
        for cpf in self._AST.domain.cpfs[1]:
            name, objects = cpf.pvar[1]
            
            # make sure the CPF is defined in pvariables {...} scope
            types = self.param_types.get(name, None)
            if types is None:
                raise RDDLUndefinedCPFError(
                    f'CPF <{name}> is not defined in pvariable block.')
            
            # make sure the number of parameters matches that in cpfs {...}
            if objects is None:
                objects = []
            if len(types) != len(objects):
                raise RDDLInvalidNumberOfArgumentsError(
                    f'l.h.s. of expression for CPF <{name}> requires '
                    f'{len(types)} parameter(s), got {objects}.')
            
            # CPFs are stored as dictionary that associates cpf name with a pair 
            # the first element is the type argument list
            # the second element is the AST expression
            objects = list(zip(objects, types))
            cpfs[name] = (objects, cpf.expr)
        
        # make sure all CPFs have a valid expression in cpfs {...}
        for (var, fluent_type) in self.variable_types.items():
            if fluent_type in {'derived-fluent', 'interm-fluent',
                               'next-state-fluent', 'observ-fluent'} \
            and var not in cpfs:
                raise RDDLMissingCPFDefinitionError(
                    f'{fluent_type} CPF <{var}> is not defined in cpfs block.')
                    
        self.cpfs = cpfs
    
    def _extract_constraints(self):
        self.terminals = getattr(self._AST.domain, 'terminals', [])
        self.preconditions = getattr(self._AST.domain, 'preconds', [])
        self.invariants = getattr(self._AST.domain, 'invariants', [])

    def _extract_horizon(self):
        horizon = self._AST.instance.horizon
        if not (horizon >= 0):
            raise RDDLValueOutOfRangeError(
                f'Horizon {horizon} in the instance is not >= 0.')
        self.horizon = horizon

    def _extract_max_actions(self):
        numactions = getattr(self._AST.instance, 'max_nondef_actions', 'pos-inf')
        if numactions == 'pos-inf':
            self.max_allowed_actions = sum(map(np.size, self.actions.values()))
        else:
            self.max_allowed_actions = int(numactions)

    def _extract_discount(self):
        discount = self._AST.instance.discount
        if not (0. <= discount):
            raise RDDLValueOutOfRangeError(
                f'Discount factor {discount} in the instance is not >= 0')
        self.discount = discount
    



from copy import deepcopy
import csv
import datetime
import gurobipy
import jax
from multiprocessing import get_context
import numpy as np
import os
import time
from typing import Dict, List

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Gurobi.GurobiRDDLPlanner import GurobiStraightLinePlan
from pyRDDLGym.Core.Gurobi.GurobiRDDLPlanner import GurobiOnlineController

# use a global instance
GLOBAL_ENV = gurobipy.Env()


# ===============================================================================
# 
# REPLANNING
#
# ===============================================================================
def objective_replan(T, kwargs, key, index):

    # transform hyper-parameters to natural space
    if kwargs['verbose']:
        print(f'[{index}] key={key}, T={T}...', flush=True)

    # initialize policy
    policy = GurobiOnlineController(
        rddl=deepcopy(kwargs['rddl']),
        plan=GurobiStraightLinePlan(**kwargs['plan_kwargs']),
        env=GLOBAL_ENV,
        rollout_horizon=T,
        **kwargs['planner_kwargs'])

    # initialize env for evaluation (need fresh copy to avoid concurrency)
    env = RDDLEnv(domain=kwargs['domain'],
                  instance=kwargs['instance'],
                  enforce_action_constraints=False)
    
    # perform training
    average_reward = 0.0
    for trial in range(kwargs['eval_trials']):
        key, subkey = jax.random.split(key)
        total_reward = policy.evaluate(env, seed=np.array(subkey)[0])['mean']
        if kwargs['verbose']:
            print(f'    [{index}] trial {trial + 1} key={subkey}, '
                  f'reward={total_reward}', flush=True)
        average_reward += total_reward / kwargs['eval_trials']        
    if kwargs['verbose']:
        print(f'[{index}] average reward={average_reward}', flush=True)
    del policy.env
    return average_reward


class GurobiParameterTuningReplan:
    
    def __init__(self, env: RDDLEnv,
                 lookahead_range: List[int]=list(range(1, 41)),
                 timeout_training: float=None,
                 timeout_tuning: float=np.inf,
                 eval_trials: int=5,
                 verbose: bool=True,
                 planner_kwargs: Dict={
                     'model_params': {'NonConvex': 2, 'OutputFlag': 0},
                     'verbose': 0,
                 },
                 plan_kwargs: Dict={},
                 pool_context: str='spawn',
                 num_workers: int=4,
                 poll_frequency: float=0.2) -> None:
        '''Creates a new instance for tuning the lookahead horizon for Gurobi
        planners.
        
        :param env: the RDDLEnv describing the MDP to optimize
        :param lookahead_range: list of lookahead horizons to validate from
        :param timeout_training: the maximum amount of time to spend training per
        trial/decision step (in seconds)
        :param timeout_tuning: the maximum amount of time to spend tuning 
        hyperparameters in general (in seconds)
        :param eval_trials: how many trials to perform independent training
        in order to estimate the return for each set of hyper-parameters
        :param verbose: whether to print intermediate results of tuning
        :param planner_kwargs: additional arguments to feed to the planner
        :param plan_kwargs: additional arguments to feed to the plan/policy
        :param pool_context: context for multiprocessing pool (defaults to 
        "spawn")
        :param num_workers: how many points to evaluate in parallel
        :param poll_frequency: how often (in seconds) to poll for completed
        jobs, necessary if num_workers > 1
        '''
        # timeout for training must be set in Gurobi variables
        if timeout_training is not None:
            if 'model_params' not in planner_kwargs:
                planner_kwargs['model_params'] = {}
            planner_kwargs['model_params']['TimeLimit'] = timeout_training
        
        self.env = env
        self.lookahead_range = [t for t in lookahead_range 
                                if t <= self.env.horizon]
        self.timeout_training = timeout_training
        self.timeout_tuning = timeout_tuning
        self.eval_trials = eval_trials
        self.verbose = verbose
        self.planner_kwargs = planner_kwargs
        self.plan_kwargs = plan_kwargs
        self.pool_context = pool_context
        self.num_workers = num_workers
        self.poll_frequency = poll_frequency
        
    def summarize_hyperparameters(self):
        print(f'hyperparameter optimizer parameters:\n'
              f'    tuned_hyper_parameters    =horizon\n'
              f'    tuning_batch_size         ={self.num_workers}\n'
              f'    tuning_timeout            ={self.timeout_tuning}\n'
              f'    mp_pool_context_type      ={self.pool_context}\n'
              f'    mp_pool_poll_frequency    ={self.poll_frequency}\n'
              f'meta-objective parameters:\n'
              f'    planning_trials_per_iter  ={self.eval_trials}\n'
              f'    planning_timeout_per_trial={self.timeout_training}')
        
    def _pickleable_objective_with_kwargs(self):
        objective_fn = objective_replan
            
        # duplicate planner and plan keyword arguments must be removed
        plan_kwargs = self.plan_kwargs.copy()        
        planner_kwargs = self.planner_kwargs.copy()
        planner_kwargs.pop('rollout_horizon', None)
                        
        kwargs = {
            'rddl': self.env.model,
            'domain': self.env.domain_text,
            'instance': self.env.instance_text,
            'planner_kwargs': planner_kwargs,
            'plan_kwargs': plan_kwargs,
            'verbose': self.verbose,
            'eval_trials': self.eval_trials
        }
        return objective_fn, kwargs
    
    @staticmethod
    def _wrapped_evaluate(index, T, key, func, kwargs):
        target = func(T=T, kwargs=kwargs, key=key, index=index)
        pid = os.getpid()
        return index, pid, T, target
    
    def tune(self, key: int, filename: str) -> Dict[str, object]:
        '''Tunes the hyper-parameters for Jax planner, returns the best found.'''
        key = jax.random.PRNGKey(key)
        self.summarize_hyperparameters()
        
        start_time = time.time()
        
        # objective function
        objective = self._pickleable_objective_with_kwargs()
        evaluate = GurobiParameterTuningReplan._wrapped_evaluate
        
        # clear and prepare output file
        filename = self._filename(filename, 'csv')
        with open(filename, 'w', newline='') as file:
            writer = csv.writer(file)
            writer.writerow(
                ['pid', 'worker', 'iteration', 'target', 'best_target', 'T']
            )
                
        # start multiprocess evaluation
        best_T, best_target = None, -np.inf
        
        for (it, i) in enumerate(range(
            0, len(self.lookahead_range), self.num_workers)):
            batch = self.lookahead_range[i:i + self.num_workers]
            
            # check if there is enough time left for another iteration
            elapsed = time.time() - start_time
            if elapsed >= self.timeout_tuning:
                print(f'global time limit reached at iteration {it}, aborting')
                break
            
            # continue with next iteration
            print('\n' + '*' * 25 + 
                  '\n' + f'[{datetime.timedelta(seconds=elapsed)}] ' + 
                  f'starting iteration {it} with batch {batch}' + 
                  '\n' + '*' * 25)
            worker_ids = list(range(len(batch)))
            key, *subkeys = jax.random.split(key, num=len(batch) + 1)
            rows = [None] * len(batch)
            
            # create worker pool: note each iteration must wait for all workers
            # to finish before moving to the next            
            with get_context(self.pool_context).Pool(processes=len(batch)) as pool:
                
                # assign jobs to worker pool
                results = [
                    pool.apply_async(evaluate, worker_args + objective)
                    for worker_args in zip(worker_ids, batch, subkeys)
                ]
            
                # wait for all workers to complete
                while results:
                    time.sleep(self.poll_frequency)
                    
                    # determine which jobs have completed
                    jobs_done = []
                    for (i, candidate) in enumerate(results):
                        if candidate.ready():
                            jobs_done.append(i)
                    
                    # get result from completed jobs
                    for i in jobs_done[::-1]:
                        index, pid, T, target = results.pop(i).get()
                        if target > best_target:
                            best_T, best_target = T, target
                        rows[index] = [pid, index, it, target, best_target, T]
                        
            # write results of all processes in current iteration to file
            with open(filename, 'a', newline='') as file:
                writer = csv.writer(file)
                writer.writerows(rows)
        
        # print summary of results
        elapsed = time.time() - start_time
        print(f'summary of hyper-parameter optimization:\n'
              f'    time_elapsed       ={datetime.timedelta(seconds=elapsed)}\n'
              f'    iterations         ={it + 1}\n'
              f'    best_T             ={best_T}\n'
              f'    best_meta_objective={best_target}\n')
        
        return best_T

    def _filename(self, name, ext):
        domainName = self.env.model.domainName()
        instName = self.env.model.instanceName()
        domainName = ''.join(c for c in domainName if c.isalnum() or c == '_')
        instName = ''.join(c for c in instName if c.isalnum() or c == '_')
        filename = f'{name}_{domainName}_{instName}.{ext}'
        return filename
    



# This file is based on thiago pbueno's pyrddl.
# https://github.com/thiagopbueno/pyrddl
# it was adapted and extended for pyRDDLGym

from typing import List, Optional, Union

FluentValue = Union[bool, int, float]


class PVariable(object):
    '''Parameterized Variable.
    Note:
        This class is intended to be solely used by the parser and compiler.
        Do not attempt to directly use this class to build a PVariable object.
    Args:
        name (str): Name of fluent.
        fluent_type (str): Type of fluent.
        range (str): Range of fluent.
        param_types (Optional[List[str]]): List of parameter types.
        default (Optional[FluentValue]): Default value of fluent.
        level (Optional[int]): Level of intermediate fluent.
    Attributes:
        name (str): Name of fluent.
        fluent_type (str): Type of fluent.
        range (str): Range of fluent.
        param_types (Optional[List[str]]): List of parameter types.
        default (Optional[FluentValue]): Default value of fluent.
        level (Optional[int]): Level of intermediate fluent.
    '''

    def __init__(self,
            name: str,
            fluent_type: str,
            range_type: str,
            param_types: Optional[List[str]] = None,
            default: Optional[FluentValue] = None,
            level: Optional[int] = None) -> None:
        self.name = name
        self.fluent_type = fluent_type
        self.range = range_type
        self.param_types = param_types
        self.default = default
        self.level = level

    @property
    def arity(self) -> int:
        '''Returns arity of fluent.'''
        return len(self.param_types) if self.param_types is not None else 0

    def is_non_fluent(self) -> bool:
        '''Returns True if fluent is of non-fluent type. False, otherwise.'''
        return self.fluent_type == 'non-fluent'

    def is_fluent(self) -> bool:
        '''Returns True if fluent is not a non-fluent type. False, otherwise.'''
        return self.fluent_type != 'non-fluent'

    def is_state_fluent(self) -> bool:
        '''Returns True if fluent is of state-fluent type. False, otherwise.'''
        return self.fluent_type == 'state-fluent'

    def is_action_fluent(self) -> bool:
        '''Returns True if fluent is of action-fluent type. False, otherwise.'''
        return self.fluent_type == 'action-fluent'

    def is_intermediate_fluent(self) -> bool:
        '''Returns True if fluent is of interm-fluent type. False, otherwise.'''
        return self.fluent_type == 'interm-fluent'

    def is_derived_fluent(self) -> bool:
        '''Returns True if fluent is of derived-fluent type. False, otherwise.'''
        return self.fluent_type == 'derived-fluent'

    def is_observ_fluent(self) -> bool:
        '''Returns True if fluent is of observ-fluent type. False, otherwise.'''
        return self.fluent_type == 'observ-fluent'

    def __str__(self) -> str:
        '''Returns string value of PVariable.'''
        return '{}/{}'.format(self.name, self.arity)

    def __repr__(self) -> str:
        '''Returns canonical string representation of PVariable.'''
        return self.name if self.arity == 0 else '{}({})'.format(self.name, ','.join(self.param_types))



import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class RaceCarInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return 'RaceCar'
    
    def get_domain_name(self) -> str:
        return 'racecar'
    
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        
        # generate random obstacles as boxes
        # two additional rectangles are generated to bound origin and goal
        boundaries = self._generate_rectangles(params['num_blocks'] + 2,
                                               params['min_block_size'],
                                               params['max_block_size'])
        boundaries.insert(0, (0., 0., 1., 1.))
        boundaries = [tuple(xi * params['scale'] for xi in pt) for pt in boundaries]
        nonfluents = {}
        objects = []
        i = 0
        for boundary in boundaries[:-2]:
            for (X1, Y1, X2, Y2) in self._line_segments(*boundary):
                obj = f'b{i + 1}'
                objects.append(obj)
                nonfluents[f'X1({obj})'] = X1
                nonfluents[f'Y1({obj})'] = Y1
                nonfluents[f'X2({obj})'] = X2
                nonfluents[f'Y2({obj})'] = Y2
                i += 1
        
        # generate state at random from last two boxes
        states = {}
        x1, y1, x2, y2 = boundaries[-2]
        X0 = random.uniform(x1, x2)
        Y0 = random.uniform(y1, y2)
        x1, y1, x2, y2 = boundaries[-1]
        GX = random.uniform(x1, x2)
        GY = random.uniform(y1, y2)
        states['x'] = nonfluents['X0'] = X0
        states['y'] = nonfluents['Y0'] = Y0
        nonfluents['GX'] = GX
        nonfluents['GY'] = GY
        nonfluents['RADIUS'] = params['goal_radius']
        states['vx'] = 0.0
        states['vy'] = 0.0
        
        return {
            'objects': {'b': objects},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }
            
    def _line_segments(self, x1, y1, x2, y2):
        return [(x1, y1, x2, y1), (x2, y1, x2, y2),
                (x2, y2, x1, y2), (x1, y2, x1, y1)]

    def _generate_rectangles(self, n, minw, maxw, dist_to_goal=0.7):
        rectangles = []
        for i in range(n):
            while True:
                width = random.uniform(minw, maxw)
                height = random.uniform(minw, maxw)
                x = random.uniform(0, 1 - width)
                y = random.uniform(0, 1 - height)
                new_rect = (x, y, x + width, y + height)
                overlap = any(self._intersect(rect, new_rect) for rect in rectangles)
                if i == n - 1:
                    x0, y0, *_ = rectangles[-1]
                    dist2 = (x0 - x) ** 2 + (y0 - y) ** 2
                    if dist2 < dist_to_goal ** 2:
                        overlap = True
                if not overlap:
                    rectangles.append(new_rect)
                    break
        return rectangles

    def _intersect(self, rect1, rect2):
        return not (rect1[2] <= rect2[0] or  # left edge of rect1 is to the right of right edge of rect2
                    rect1[0] >= rect2[2] or  # right edge of rect1 is to the left of left edge of rect2
                    rect1[3] <= rect2[1] or  # top edge of rect1 is below bottom edge of rect2
                    rect1[1] >= rect2[3])  # bottom edge of rect1 is above top edge of rect2


